from .Interface import Interface, VsxOlr2CaptureInformationStructure, VsxOlr2ModbusDataStructure
from .Interface import VsxDataContainerHandle, VsxTagList, VsxImage, VsxCaptureInformationStructure
from .Interface import VsxTransformationStructure, VsxDisparityDescriptor2Structure, VsxLineData
from .Interface import VsxLineCoordinateStructure
from typing import Tuple, Optional, List
from ctypes import *
import numpy as np
from enum import IntEnum


class ImageData2Format(IntEnum):
    VSX_IMAGE_DATA2_FORMAT_MONO8 = 17301505
    VSX_IMAGE_DATA2_FORMAT_CONFIDENCE8 = 17301702
    VSX_IMAGE_DATA2_FORMAT_MONO12 = 17825797
    VSX_IMAGE_DATA2_FORMAT_MONO16 = 17825799
    VSX_IMAGE_DATA2_FORMAT_COORD3D_A16 = 17825974
    VSX_IMAGE_DATA2_FORMAT_COORD3D_B16 = 17825975
    VSX_IMAGE_DATA2_FORMAT_COORD3D_C16 = 17825976
    VSX_IMAGE_DATA2_FORMAT_COORD3D_A32f = 18874557
    VSX_IMAGE_DATA2_FORMAT_COORD3D_B32f = 18874558
    VSX_IMAGE_DATA2_FORMAT_COORD3D_C32f = 18874559


class VsxCaptureInformation:
    """
    CaptureInformation contains information identifying and describing the captured image. The timestamp reflects a monotonic clock,
    without further synchronization timestamps are only meaningful relative to one another.
    The frame counter counts the number of actually captured images, while the trigger counter also includes missed trigger events
    (if they could be detected at all).
    The JobId and ParameterId jointly specify the parameters that were used for evaluating the image (including disparity computation).
    They do not necessarily reflect any settings that affect the actual image capturing. These are given separately instead.
    The reason for that behavior is that capture settings updates are hard to do atomically if you do not want to disable triggers
    (which would affect latency and in the worst case miss triggers).
    """
    def __init__(self):
        self.trigger_counter: int = 0
        self.parameter_id: int = 0
        self.job_id: int = 0
        self.rotary_encoder: int = 0
        self.frame_counter: int = 0
        self.timestamp: int = 0
        self.exposure_time: int = 0
        self.gain: int = 0
        self.illumination: int = 0
        self.trigger_source: int = 0


class VsxLineCoordinate:
    def __init__(self, c: float, x: float, y: float, z: float, q: float, i: float):
        self.c: float = c
        self.x: float = x
        self.y: float = y
        self.z: float = z
        self.q: float = q
        self.i: float = i


class VsxTransformation:
    """
    Used to transform raw point cloud data from device.
    """
    def __init__(self):
        self.translation_tx: float = float('NaN')
        self.translation_ty: float = float('NaN')
        self.translation_tz: float = float('NaN')
        self.quaternion_q0: float = float('NaN')
        self.quaternion_q1: float = float('NaN')
        self.quaternion_q2: float = float('NaN')
        self.quaternion_q3: float = float('NaN')


class VsxDisparityDescriptor2:
    def __init__(self):
        self.focal_length: float = float('NaN')
        self.principal_point_u: float = float('NaN')
        self.principal_point_v: float = float('NaN')
        self.baseline: float = float('NaN')
        self.offset_left_rectified_to_disparity_u: float = float('NaN')
        self.offset_left_rectified_to_disparity_v: float = float('NaN')


class VsxOlr2ModbusData :
    def __init__(self):
        self.activation_timer: int = 0
        self.compare_buffer: int = 0
        self.target_position: int = 0
        self.robot_data: list[int] = [None] * 13


class VsxOlr2CaptureInformation:
    def __init__(self):
        self.frame_counter: int = 0
        self.trigger_counter: int = 0
        self.current_position: float = 0
        self.io_state: int = 0
        self.timestamp: int = 0
        self.lma_exposure_time_1: int = 0
        self.lma_exposure_time_2: int = 0
        self.lmb_exposure_time_1: int = 0
        self.lmb_exposure_time_2: int = 0
        self.lma_roi_offset_x: int = 0
        self.lma_roi_length_x: int = 0
        self.lma_roi_offset_z: int = 0
        self.lma_roi_length_z: int = 0
        self.lmb_roi_offset_x: int = 0
        self.lmb_roi_length_x: int = 0
        self.lmb_roi_offset_z: int = 0
        self.lmb_roi_length_z: int = 0
        self.auto_trigger_frame_rate: int = 0
        self.trigger_source: int = 0


class DataContainer:
    def __init__(self, *, _is_direct=True):
        # don't initialize me directly
        if _is_direct:
            raise TypeError("create with Sensor.GetDataContainer*")
        self.data_container_handle = POINTER(VsxDataContainerHandle)()

    def __del__(self):
        _res = Interface.release_data_container(self.data_container_handle)
        self.data_container_handle = None

    def SaveData(self, tag: str, file_name: str) -> int:
        """
        Saves a VsxMessage to the given filename.
        @param tag: Specify which tag from container should be saved ("*" save complete container)
        @param file_name: Path and filename where to save the message
        @return: Returns VSX_STATUS_SUCCESS(0) on success
        """
        return Interface.save_data(self.data_container_handle, tag.encode("utf-8"), file_name.encode("utf-8"))

    def Save3DPointCloudData(self, tag_x: str, tag_y: str, tag_z: str, file_name: str) -> int:
        """
        Saves a 3D point cloud as pcd to the given filename.
        @param tag_x: The x image tag name
        @param tag_y: The y image tag name
        @param tag_z: The z image tag name
        @param file_name: Path and filename where to save the data
        @return: Returns VSX_STATUS_SUCCESS(0) on success
        """
        return Interface.save_3d_point_cloud_data(self.data_container_handle, tag_x.encode("utf-8"),
                                                  tag_y.encode("utf-8"),
                                                  tag_z.encode("utf-8"), file_name.encode("utf-8"))

    def GetTagList(self):
        """
        Returns all available tags from a dynamic container
        @return: Returns VSX_STATUS_SUCCESS(0) on success and tag list object
        """
        result_ptr = POINTER(VsxTagList)()
        ret = Interface.get_tag_list(self.data_container_handle, byref(result_ptr))
        if ret == 0:
            results_list = []
            for i in range(result_ptr.contents.length):
                results_list.append(result_ptr.contents.tags[i].decode("utf-8"))

            ret = Interface.release_tag_list(byref(result_ptr))
            return ret, results_list
        else:
            return ret, None

    def GetCaptureInformation(self, tag: str) -> Tuple[int, Optional[VsxCaptureInformation]]:
        """
        Get capture information from a dynamic container
        @param tag: Tag name of data
        @return: Returns VSX_STATUS_SUCCESS(0) on success, VsxCaptureInformation object
        """
        result_ptr = POINTER(VsxCaptureInformationStructure)()
        ret = Interface.get_capture_information(self.data_container_handle, tag.encode("utf-8"), byref(result_ptr))
        if ret == 0:
            results = VsxCaptureInformation()
            results.trigger_counter = result_ptr.contents.triggerCounter
            results.parameter_id = result_ptr.contents.parameterId
            results.job_id = result_ptr.contents.jobId
            results.rotary_encoder = result_ptr.contents.rotaryEncoder
            results.frame_counter = result_ptr.contents.frameCounter
            results.timestamp = result_ptr.contents.timestamp
            results.exposure_time = result_ptr.contents.exposureTime
            results.gain = result_ptr.contents.gain
            results.illumination = result_ptr.contents.illumination
            results.trigger_source = result_ptr.contents.triggerSource
            ret = Interface.release_capture_information(byref(result_ptr))
            return ret, results
        else:
            return ret, None

    def GetTransformation(self, tag: str) -> Tuple[int, Optional[VsxTransformation]]:
        """
        Get transformation from a dynamic container
        @param tag: Tag name of data
        @return: Returns VSX_STATUS_SUCCESS(0) on success, transformation object
        """
        result_ptr = POINTER(VsxTransformationStructure)()
        ret = Interface.get_transformation(self.data_container_handle, tag.encode("utf-8"), byref(result_ptr))
        if ret == 0:
            results = VsxTransformation()
            results.translation_tx = result_ptr.contents.translationTX
            results.translation_ty = result_ptr.contents.translationTY
            results.translation_tz = result_ptr.contents.translationTZ
            results.quaternion_q0 = result_ptr.contents.quaternionQ0
            results.quaternion_q1 = result_ptr.contents.quaternionQ1
            results.quaternion_q2 = result_ptr.contents.quaternionQ2
            results.quaternion_q3 = result_ptr.contents.quaternionQ3
            ret = Interface.release_transformation(byref(result_ptr))
            return ret, results
        else:
            return ret, None

    def GetDisparityDescriptor2(self, tag: str) -> Tuple[int, Optional[VsxDisparityDescriptor2]]:
        """
        Get disparity descriptor from a dynamic container
        @param tag: Tag name of data
        @return: Returns VSX_STATUS_SUCCESS(0) on success, VsxDisparityDescriptor2 object
        """
        result_ptr = POINTER(VsxDisparityDescriptor2Structure)()
        ret = Interface.get_disparity_descriptor2(self.data_container_handle, tag.encode("utf-8"), byref(result_ptr))
        if ret == 0:
            results = VsxDisparityDescriptor2()
            results.focal_length = result_ptr.contents.focalLength
            results.principal_point_u = result_ptr.contents.principalPointU
            results.principal_point_v = result_ptr.contents.principalPointV
            results.baseline = result_ptr.contents.baseline
            results.offset_left_rectified_to_disparity_u = result_ptr.contents.offsetLeftRectifiedToDisparityU
            results.offset_left_rectified_to_disparity_v = result_ptr.contents.offsetLeftRectifiedToDisparityV
            ret = Interface.release_disparity_descriptor2(byref(result_ptr))
            return ret, results
        else:
            return ret, None

    def GetOlr2CaptureInformation(self, tag: str) -> Tuple[int, Optional[VsxOlr2CaptureInformation]]:
        """
        Get olr2 capture information from a dynamic container
        @param tag: Tag name of data
        @return: Returns VSX_STATUS_SUCCESS(0) on success, VsxOlr2CaptureInformation object
        """
        result_ptr = POINTER(VsxOlr2CaptureInformationStructure)()
        ret = Interface.get_olr2_capture_information(self.data_container_handle, tag.encode("utf-8"), byref(result_ptr))
        if ret == 0:
            results = VsxOlr2CaptureInformation()
            results.frame_counter = result_ptr.contents.frameCounter
            results.trigger_counter = result_ptr.contents.triggerCounter
            results.current_position = result_ptr.contents.currentPosition
            results.io_state = result_ptr.contents.ioState
            results.timestamp = result_ptr.contents.timestamp
            results.lma_exposure_time_1 = result_ptr.contents.lmaExposureTime1
            results.lma_exposure_time_2 = result_ptr.contents.lmaExposureTime2
            results.lmb_exposure_time_1 = result_ptr.contents.lmbExposureTime1
            results.lmb_exposure_time_2 = result_ptr.contents.lmbExposureTime2
            results.lma_roi_offset_x = result_ptr.contents.lmaRoiOffsetX
            results.lma_roi_length_x = result_ptr.contents.lmaRoiLengthX
            results.lma_roi_offset_z = result_ptr.contents.lmaRoiOffsetZ
            results.lma_roi_length_z = result_ptr.contents.lmaRoiLengthZ
            results.lmb_roi_offset_x = result_ptr.contents.lmbRoiOffsetX
            results.lmb_roi_length_x = result_ptr.contents.lmbRoiLengthX
            results.lmb_roi_offset_z = result_ptr.contents.lmbRoiOffsetZ
            results.lmb_roi_length_z = result_ptr.contents.lmbRoiLengthZ
            results.auto_trigger_frame_rate = result_ptr.contents.autoTriggerFrameRate
            results.trigger_source = result_ptr.contents.triggerSource

            ret = Interface.release_olr2_capture_information(byref(result_ptr))
            return ret, results
        else:
            return ret, None

    def GetOlr2ModbusData(self, tag: str) -> Tuple[int, Optional[VsxOlr2ModbusData]]:
        """
        Get modbus data for olr2 sensor from a dynamic container
        @param tag: Tag name of data
        @return: Returns VSX_STATUS_SUCCESS(0) on success, VsxOlr2ModbusData object
        """
        result_ptr = POINTER(VsxOlr2ModbusDataStructure)()
        ret = Interface.get_olr2_modbus_data(self.data_container_handle, tag.encode("utf-8"), byref(result_ptr))
        if ret == 0:
            results = VsxOlr2ModbusData()
            results.activation_timer = result_ptr.contents.activationTimer
            results.compare_buffer = result_ptr.contents.compareBuffer
            results.target_position = result_ptr.contents.targetPosition
            results.robot_data = result_ptr.contents.robotData
            ret = Interface.release_olr2_modbus_data(byref(result_ptr))
            return ret, results
        else:
            return ret, None

    def GetImage(self, image_tag: str) -> Tuple[int, Optional[np.ndarray], Optional[dict]]:
        """
        Get image from a dynamic container, access via numpy array
        @param image_tag: Tag name of image data
        @return: Returns VSX_STATUS_SUCCESS(0) on success, numpy array, dictionary of attributes
        """
        result_ptr = POINTER(VsxImage)()
        ret = Interface.get_image(self.data_container_handle, image_tag.encode("utf-8"), byref(result_ptr))
        if ret == 0:
            format_image = ImageData2Format(result_ptr.contents.format)
            if format_image == ImageData2Format.VSX_IMAGE_DATA2_FORMAT_MONO8 \
                    or format_image == ImageData2Format.VSX_IMAGE_DATA2_FORMAT_CONFIDENCE8:
                format_type = c_uint8
            elif format_image == ImageData2Format.VSX_IMAGE_DATA2_FORMAT_MONO12 \
                    or format_image == ImageData2Format.VSX_IMAGE_DATA2_FORMAT_MONO16:
                format_type = c_uint16
            elif format_image == ImageData2Format.VSX_IMAGE_DATA2_FORMAT_COORD3D_A16 \
                    or format_image == ImageData2Format.VSX_IMAGE_DATA2_FORMAT_COORD3D_B16 \
                    or format_image == ImageData2Format.VSX_IMAGE_DATA2_FORMAT_COORD3D_C16:
                format_type = c_int16
            elif format_image == ImageData2Format.VSX_IMAGE_DATA2_FORMAT_COORD3D_A32f \
                    or format_image == ImageData2Format.VSX_IMAGE_DATA2_FORMAT_COORD3D_B32f \
                    or format_image == ImageData2Format.VSX_IMAGE_DATA2_FORMAT_COORD3D_C32f:
                format_type = c_float
            else:
                assert None, "Unknown format"
                return -1, None, None

            data_pointer = cast(result_ptr.contents.raw_data, POINTER(format_type))
            img_array = np.ctypeslib.as_array(data_pointer,
                                              shape=(result_ptr.contents.height, result_ptr.contents.width))
            image = np.copy(img_array)
            results = {'format': format_image,
                       'width': result_ptr.contents.width,
                       'height': result_ptr.contents.height,
                       'line_pitch': result_ptr.contents.line_pitch,
                       'frame_counter': result_ptr.contents.frame_counter,
                       'coordinate_scale': result_ptr.contents.coordinate_scale,
                       'coordinate_offset': result_ptr.contents.coordinate_offset,
                       'axis_min': result_ptr.contents.axis_min,
                       'axis_max': result_ptr.contents.axis_max,
                       'invalid_data_value': result_ptr.contents.invalid_data_value,
                       }

            del img_array
            ret = Interface.release_image(byref(result_ptr))
            return ret, image, results
        else:
            return ret, None, None

    def GetLine(self, image_tag: str) -> Tuple[int, Optional[List[List[VsxLineCoordinate]]], Optional[dict]]:
        """
        Get line data from a dynamic container
        @param image_tag: Tag name of line data
        @return: Returns VSX_STATUS_SUCCESS(0) on success, list of coordinates, dictionary of attributes
        """
        result_ptr = POINTER(VsxLineData)()
        ret = Interface.get_line(self.data_container_handle, image_tag.encode("utf-8"), byref(result_ptr))
        if ret == 0:
            data_pointer = cast(result_ptr.contents.lines, POINTER(POINTER(VsxLineCoordinateStructure)))
            lines = [[VsxLineCoordinate(data_pointer[i][j].c, data_pointer[i][j].x, data_pointer[i][j].y,
                                        data_pointer[i][j].z, data_pointer[i][j].q, data_pointer[i][j].i)
                      for j in range(0, result_ptr.contents.width)] for i in range(0, result_ptr.contents.countLines)]

            results = {'format': result_ptr.contents.format,
                       'width': result_ptr.contents.width,
                       'count_lines': result_ptr.contents.countLines,
                       'frame_counter': result_ptr.contents.frameCounter,
                       'min_x': result_ptr.contents.minX,
                       'max_x': result_ptr.contents.maxX,
                       'min_z': result_ptr.contents.minZ,
                       'max_z': result_ptr.contents.maxZ,
                       }

            ret = Interface.release_line(byref(result_ptr))
            return ret, lines, results
        else:
            return ret, None, None

    def GetResultXml(self, tag_result: str) -> Tuple[int, Optional[str]]:
        """
        Returns the complete xml response from an result inside data container
        @param tag_result: Name of result
        @return: Returns VSX_STATUS_SUCCESS(0) on success, complete result xml as string
        """
        result_ptr = c_char_p()
        ret = Interface.get_result_xml(self.data_container_handle, tag_result.encode("utf-8"), result_ptr)
        if ret == 0:
            str_result = result_ptr.value.decode("utf-8")
            ret = Interface.release_string(result_ptr)
            return ret, str_result
        else:
            return ret, None

    def GetResultElementString(self, tag_result: str, xpath: str) -> Tuple[int, Optional[str]]:
        """
        Return certain value from a result inside data container
        @param tag_result: Name of result
        @param xpath: xPath definition
        @return: Returns VSX_STATUS_SUCCESS(0) on success, result as string
        """
        result_ptr = c_char_p()
        ret = Interface.get_result_element_string(self.data_container_handle, tag_result.encode("utf-8"),
                                                  xpath.encode("utf-8"), result_ptr)
        if ret == 0:
            str_result = result_ptr.value.decode("utf-8")
            ret = Interface.release_string(result_ptr)
            return ret, str_result
        else:
            return ret, None

    def GetResultElementInt32(self, tag_result: str, xpath: str) -> Tuple[int, Optional[int]]:
        """
        Return certain value from a result inside data container
        @param tag_result: Name of result
        @param xpath: xPath definition
        @return: Returns VSX_STATUS_SUCCESS(0) on success, result as integer number
        """
        result_ptr = c_int32()
        ret = Interface.get_result_element_int32(self.data_container_handle, tag_result.encode("utf-8"),
                                                 xpath.encode("utf-8"), result_ptr)
        if ret == 0:
            return ret, result_ptr.value
        else:
            return ret, None

    def GetResultElementInt64(self, tag_result: str, xpath: str) -> Tuple[int, Optional[int]]:
        """
        Return certain value from a result inside data container
        @param tag_result: Name of result
        @param xpath: xPath definition
        @return: Returns VSX_STATUS_SUCCESS(0) on success, result as integer number
        """
        result_ptr = c_int64()
        ret = Interface.get_result_element_int64(self.data_container_handle, tag_result.encode("utf-8"),
                                                 xpath.encode("utf-8"), result_ptr)
        if ret == 0:
            return ret, result_ptr.value
        else:
            return ret, None

    def GetResultElementDouble(self, tag_result: str, xpath: str) -> Tuple[int, Optional[float]]:
        """
        Return certain value from a result inside data container
        @param tag_result: Name of result
        @param xpath: xPath definition
        @return: Returns VSX_STATUS_SUCCESS(0) on success, result as float number
        """
        result_ptr = c_double()
        ret = Interface.get_result_element_double(self.data_container_handle, tag_result.encode("utf-8"),
                                                  xpath.encode("utf-8"), result_ptr)
        if ret == 0:
            return ret, result_ptr.value
        else:
            return ret, None
