from ctypes import *
from pathlib import Path
from platform import architecture, system, machine
import os


class VsxSystemHandle(Structure):
    _fields_ = [('handle', c_int32)]

    def __repr__(self):
        return '(handle {0})'.format(self.handle)


class VsxDataContainerHandle(Structure):
    _fields_ = [('handle', c_int32)]

    def __repr__(self):
        return '(handle {0})'.format(self.handle)


class VsxTagList(Structure):
    _fields_ = [('length', c_int32),
                ('tags', POINTER(c_char_p)),
                ]


class VsxImage(Structure):
    _fields_ = [('raw_data', c_void_p),
                ('format', c_int32),
                ('width', c_int32),
                ('height', c_int32),
                ('line_pitch', c_int32),
                ('frame_counter', c_long),
                ('coordinate_scale', c_double),
                ('coordinate_offset', c_double),
                ('axis_min', c_double),
                ('axis_max', c_double),
                ('invalid_data_value', c_double),
                ]


class VsxLineCoordinateStructure(Structure):
    _fields_ = [('c', c_float),
                ('x', c_float),
                ('y', c_float),
                ('z', c_float),
                ('q', c_float),
                ('i', c_float),
                ]


class VsxLineData(Structure):
    _fields_ = [('lines', c_void_p),
                ('format', c_uint16),
                ('width', c_uint16),
                ('countLines', c_uint16),
                ('frameCounter', c_uint16),
                ('minX', c_float),
                ('maxX', c_float),
                ('minZ', c_float),
                ('maxZ', c_float),
                ]


class VsxCaptureInformationStructure(Structure):
    _fields_ = [('triggerCounter', c_uint64),
                ('parameterId', c_uint64),
                ('jobId', c_uint64),
                ('rotaryEncoder', c_int64),
                ('frameCounter', c_uint64),
                ('timestamp', c_uint64),
                ('exposureTime', c_uint32),
                ('gain', c_uint32),
                ('illumination', c_uint8),
                ('triggerSource', c_uint8),
                ]


class VsxDisparityDescriptor2Structure(Structure):
    _fields_ = [('focalLength', c_double),
                ('principalPointU', c_double),
                ('principalPointV', c_double),
                ('baseline', c_double),
                ('offsetLeftRectifiedToDisparityU', c_double),
                ('offsetLeftRectifiedToDisparityV', c_double),
                ]


class VsxOlr2CaptureInformationStructure(Structure):
    _fields_ = [('frameCounter', c_uint64),
                ('triggerCounter', c_uint64),
                ('currentPosition', c_double),
                ('ioState', c_uint64),
                ('timestamp', c_uint64),
                ('lmaExposureTime1', c_uint32),
                ('lmaExposureTime2', c_uint32),
                ('lmbExposureTime1', c_uint32),
                ('lmbExposureTime2', c_uint32),
                ('lmaRoiOffsetX', c_uint16),
                ('lmaRoiLengthX', c_uint16),
                ('lmaRoiOffsetZ', c_uint16),
                ('lmaRoiLengthZ', c_uint16),
                ('lmbRoiOffsetX', c_uint16),
                ('lmbRoiLengthX', c_uint16),
                ('lmbRoiOffsetZ', c_uint16),
                ('lmbRoiLengthZ', c_uint16),
                ('autoTriggerFrameRate', c_uint16),
                ('triggerSource', c_uint8),
                ]


class VsxOlr2ModbusDataStructure(Structure):
    _fields_ = [('activationTimer', c_uint16),
                ('compareBuffer', c_uint16),
                ('targetPosition', c_uint16),
                ('robotData', c_uint16 * 13),
                ]


class VsxTransformationStructure(Structure):
    _fields_ = [('translationTX', c_double),
                ('translationTY', c_double),
                ('translationTZ', c_double),
                ('quaternionQ0', c_double),
                ('quaternionQ1', c_double),
                ('quaternionQ2', c_double),
                ('quaternionQ3', c_double),
                ]


class VsxDevice(Structure):
    _fields_ = [('ipAddress', c_char_p),
                ('networkMask', c_char_p),
                ('gateway', c_char_p),
                ('macAddress', c_char_p),
                ('firmwareVersion', c_char_p),
                ('sensorType', c_char_p),
                ('sensorName', c_char_p),
                ('busy', c_int32),
                ('deviceVsxVersionMajor', c_int32),
                ('deviceVsxVersionMinor', c_int32),
                ('comPort', c_char_p),
                ('baudrate', c_int32),
                ('headAddress', c_char_p),
                ('isLoginNeeded', c_int32),
                ]


class VsxDeviceList(Structure):
    _fields_ = [('length', c_int32),
                ('devices', POINTER(VsxDevice)),
                ]


class VsxParameterEnumItem(Structure):
    _fields_ = [('id', c_char_p),
                ('name', c_char_p)]


class VsxParameter(Structure):
    _fields_ = [('valueString', c_char_p),
                ('valueInt', c_int32),
                ('valueDouble', c_double),
                ('valueType', c_int32),
                ('name', c_char_p),
                ('parameterId', c_char_p),
                ('configId', c_char_p),
                ('configVersion', c_int32),
                ('settingsVersion', c_int32),
                ('enumItemListLength', c_int32),
                ('enumItemList', POINTER(VsxParameterEnumItem))
                ]


class VsxParameterList(Structure):
    _fields_ = [('length', c_int32),
                ('parameters', POINTER(VsxParameter)),
                ]

    def __init__(self, length):
        super().__init__()
        self.length = length
        # Allocate an array of VsxParameter structures.
        parameters = (VsxParameter * length)()

        self.parameters = cast(parameters, POINTER(VsxParameter))


class VsxStatusItem(Structure):
    _fields_ = [('valueString', c_char_p),
                ('valueInt', c_int32),
                ('valueDouble', c_double),
                ('valueType', c_int32),
                ('name', c_char_p),
                ('statusItemId', c_char_p),
                ('configurationClass', c_char_p),
                ('configVersion', c_int32),
                ('settingsVersion', c_int32),
                ('time', c_uint64),
                ('sensorTime', c_uint64),
                ]


class VsxStatusItemList(Structure):
    _fields_ = [('length', c_int32),
                ('statusItems', POINTER(VsxStatusItem)),
                ]


class Interface(object):
    __driver = None

    if system() == 'Windows':
        callback_on_disconnect = WINFUNCTYPE(None, c_int32, c_char_p, c_int32, c_char_p)
        callback_on_session_message_received = WINFUNCTYPE(None, c_int32, c_int32, c_int32)
        callback_on_device_status_received = WINFUNCTYPE(None, c_int32, c_int32, POINTER(VsxStatusItemList))
    else:
        callback_on_disconnect = CFUNCTYPE(None, c_int32, c_char_p, c_int32, c_char_p)
        callback_on_session_message_received = CFUNCTYPE(None, c_int32, c_int32, c_int32)
        callback_on_device_status_received = CFUNCTYPE(None, c_int32, c_int32, POINTER(VsxStatusItemList))

    release_string = None
    get_library_version = None
    get_error_text = None
    init_tcp_sensor = None
    init_serial_sensor = None
    release_sensor = None
    reconnect_tcp_device = None
    reconnect_and_login_tcp_device = None
    reconnect_serial_device = None
    connect = None
    connect_ex = None
    connect_and_login = None
    connect_ex_and_login = None
    login = None
    logout = None
    set_password = None
    get_connected = None
    disconnect = None
    register_on_disconnect = None
    deregister_on_disconnect = None
    register_on_session_message_received = None
    deregister_on_session_message_received = None
    send_session_keep_alive = None
    test_system = None
    test_system_ex = None
    get_wait_timeout = None
    set_wait_timeout = None
    upload_data = None
    send_firmware = None
    send_xml_data_message = None
    set_network_settings = None
    set_network_settings_via_udp = None
    reset_dynamic_container_grabber = None
    get_data_container = None
    get_cached_container = None
    release_data_container = None
    save_data = None
    save_3d_point_cloud_data = None
    get_capture_information = None
    release_capture_information = None
    get_transformation = None
    release_transformation = None
    get_disparity_descriptor2 = None
    release_disparity_descriptor2 = None
    get_olr2_capture_information = None
    release_olr2_capture_information = None
    get_olr2_modbus_data = None
    release_olr2_modbus_data = None
    get_image = None
    release_image = None
    get_line = None
    release_line = None
    get_tag_list = None
    release_tag_list = None
    get_missing_container_frames_counter = None
    get_dynamic_container_queue_size = None
    get_number_of_cached_containers = None
    get_device_information = None
    release_device = None
    get_udp_device_list = None
    release_device_list = None
    reset_log_message_grabber = None
    get_log_message = None
    get_log_message_queue_size = None
    get_missing_log_messages_counter = None
    set_single_parameter_value = None
    set_single_parameter_value_int32 = None
    set_single_parameter_value_double = None
    get_single_parameter_value = None
    get_single_parameter_value_int32 = None
    get_single_parameter_value_double = None
    load_default_parameter_set_on_device = None
    load_parameter_set_on_device = None
    save_parameter_set_on_device = None
    upload_parameter_set = None
    download_parameter_set = None
    get_parameter_list = None
    upload_parameter_list = None
    set_single_parameter_double = None
    set_single_parameter_int32 = None
    set_single_parameter_string = None
    get_single_parameter = None
    release_parameter = None
    release_parameter_list = None
    get_result_xml = None
    get_result_element_string = None
    get_result_element_int32 = None
    get_result_element_int64 = None
    get_result_element_double = None
    get_all_device_status_data = None
    release_status_item_list = None
    register_on_device_status_received = None
    deregister_on_device_status_received = None
    subscribe_to_device_status_data = None
    unsubscribe_to_device_status_data = None

    @staticmethod
    def __datafile_path():
        # Use external library path to e.g. better debug .net libraries
        if os.getenv('VSX_PROTOCOL_DRIVER_LIBRARY') is not None:
            return os.getenv('VSX_PROTOCOL_DRIVER_LIBRARY')

        base_dir = Path(__file__).resolve().parent
        file_dir = base_dir / 'net'

        os_name = system()
        os_machine = machine()
        os_bits = architecture()[0]
        if os_name == 'Linux':
            if os_bits == '64bit':
                if os_machine == 'aarch64':
                    return str(file_dir / 'linux-arm64' / 'PF.VsxProtocolDriver.WrapperNE.so')
                elif os_machine in ("i386", "AMD64", "x86_64"):
                    return str(file_dir / 'linux-x64' / 'PF.VsxProtocolDriver.WrapperNE.so')
            else:
                if os_machine == 'armv7l' or 'aarch64':  # aarch64 needed for new raspberry os
                    return str(file_dir / 'linux-arm' / 'PF.VsxProtocolDriver.WrapperNE.so')

        if os_name == 'Windows':
            if os_bits == '64bit':
                return str(file_dir / 'win-x64' / 'PF.VsxProtocolDriver.WrapperNE.dll')
            else:
                return str(file_dir / 'win-x86' / 'PF.VsxProtocolDriver.WrapperNE.dll')
        raise OSError(f"Unknown operating system and processor combination {os_name} {os_machine} {os_bits}")

    @classmethod
    def __wrap_function(cls, func_name, restype, argtypes):
        """Simplify wrapping ctypes functions"""

        func = cls.__driver.__getattr__(func_name)
        func.restype = restype
        func.argtypes = argtypes
        return func

    @classmethod
    def init_driver(cls):
        if cls.__driver is None:
            if system() == 'Windows':
                cls.__driver = windll.LoadLibrary(Interface.__datafile_path())
            else:
                cls.__driver = cdll.LoadLibrary(Interface.__datafile_path())

            # vsx_ReleaseString()
            cls.release_string = cls.__wrap_function("vsx_ReleaseString", c_int32, [POINTER(c_char_p)])

            # vsx_GetLibraryVersion()
            cls.get_library_version = cls.__wrap_function("vsx_GetLibraryVersion", c_int32, [POINTER(c_char_p)])

            # vsx_GetErrorText()
            cls.get_error_text = cls.__wrap_function("vsx_GetErrorText", c_int32, [c_int32, POINTER(c_char_p)])

            # vsx_InitTcpSensor()
            cls.init_tcp_sensor = cls.__wrap_function("vsx_InitTcpSensor", c_int32,
                                                      [POINTER(POINTER(VsxSystemHandle)), c_char_p, c_char_p])

            # vsx_InitSerialSensor()
            cls.init_serial_sensor = cls.__wrap_function("vsx_InitSerialSensor", c_int32,
                                                         [POINTER(POINTER(VsxSystemHandle)), c_char_p, c_int32,
                                                          c_char_p, c_int32, c_char_p])

            # vsx_ReleaseSensor
            cls.release_sensor = cls.__wrap_function("vsx_ReleaseSensor", c_int32, [POINTER(POINTER(VsxSystemHandle))])

            # vsx_ReConnectTcpDevice
            cls.reconnect_tcp_device = cls.__wrap_function("vsx_ReConnectTcpDevice", c_int32,
                                                           [POINTER(VsxSystemHandle), c_char_p])

            # vsx_ReConnectAndLoginTcpDevice
            cls.reconnect_and_login_tcp_device = cls.__wrap_function("vsx_ReConnectAndLoginTcpDevice", c_int32,
                                                                     [POINTER(VsxSystemHandle), c_char_p, c_char_p,
                                                                      c_char_p])

            # vsx_ReConnectSerialDevice
            cls.reconnect_serial_device = cls.__wrap_function("vsx_ReConnectSerialDevice", c_int32,
                                                              [POINTER(VsxSystemHandle), c_char_p, c_int32, c_int32])

            # vsx_Connect
            cls.connect = cls.__wrap_function("vsx_Connect", c_int32, [POINTER(VsxSystemHandle)])

            # vsx_ConnectEx
            cls.connect_ex = cls.__wrap_function("vsx_ConnectEx", c_int32, [POINTER(VsxSystemHandle), c_int32])

            # vsx_ConnectAndLogin
            cls.connect_and_login = cls.__wrap_function("vsx_ConnectAndLogin", c_int32,
                                                        [POINTER(VsxSystemHandle), c_char_p, c_char_p])

            # vsx_ConnectExAndLogin
            cls.connect_ex_and_login = cls.__wrap_function("vsx_ConnectExAndLogin", c_int32,
                                                           [POINTER(VsxSystemHandle), c_char_p, c_char_p, c_int32])

            # vsx_Login
            cls.login = cls.__wrap_function("vsx_Login", c_int32, [POINTER(VsxSystemHandle), c_char_p, c_char_p])

            # vsx_Logout
            cls.logout = cls.__wrap_function("vsx_Logout", c_int32, [POINTER(VsxSystemHandle)])

            # vsx_SetPassword
            cls.set_password = cls.__wrap_function("vsx_SetPassword", c_int32,
                                                   [POINTER(VsxSystemHandle), c_char_p, c_char_p, c_char_p, c_char_p])

            # vsx_GetConnected
            cls.get_connected = cls.__wrap_function("vsx_GetConnected", c_int32, [POINTER(VsxSystemHandle),
                                                                                  POINTER(c_int32)])

            # vsx_Disconnect
            cls.disconnect = cls.__wrap_function("vsx_Disconnect", c_int32, [POINTER(VsxSystemHandle)])

            # vsx_RegisterOnDisconnect
            cls.register_on_disconnect = cls.__wrap_function("vsx_RegisterOnDisconnect", c_int32,
                                                             [POINTER(VsxSystemHandle), cls.callback_on_disconnect])

            # vsx_DeregisterOnDisconnect
            cls.deregister_on_disconnect = cls.__wrap_function("vsx_DeregisterOnDisconnect", c_int32,
                                                               [POINTER(VsxSystemHandle)])

            # vsx_RegisterOnSessionMessageReceived
            cls.register_on_session_message_received = cls.__wrap_function("vsx_RegisterOnSessionMessageReceived",
                                                                           c_int32,
                                                                           [POINTER(VsxSystemHandle),
                                                                            cls.callback_on_session_message_received])

            # vsx_DeregisterOnSessionMessageReceived
            cls.deregister_on_session_message_received = cls.__wrap_function("vsx_DeregisterOnSessionMessageReceived",
                                                                             c_int32,
                                                                             [POINTER(VsxSystemHandle)])

            # vsx_SendSessionKeepAlive
            cls.send_session_keep_alive = cls.__wrap_function("vsx_SendSessionKeepAlive", c_int32,
                                                              [POINTER(VsxSystemHandle)])

            # vsx_TestSystem()
            cls.test_system = cls.__wrap_function("vsx_TestSystem", c_int32,
                                                  [POINTER(VsxSystemHandle), c_char_p, c_char_p, POINTER(c_char_p),
                                                   POINTER(c_int32)])

            # vsx_TestSystem()
            cls.test_system_ex = cls.__wrap_function("vsx_TestSystemEx", c_int32,
                                                     [POINTER(VsxSystemHandle), c_char_p, c_char_p, POINTER(c_char_p),
                                                      POINTER(c_int32), c_int32])

            # vsx_GetWaitTimeout
            cls.get_wait_timeout = cls.__wrap_function("vsx_GetWaitTimeout", c_int32,
                                                       [POINTER(VsxSystemHandle), POINTER(c_int32)])

            # vsx_SetWaitTimeout
            cls.set_wait_timeout = cls.__wrap_function("vsx_SetWaitTimeout", c_int32,
                                                       [POINTER(VsxSystemHandle), c_int32])

            # vsx_UploadData()
            cls.upload_data = cls.__wrap_function("vsx_UploadData", c_int32, [POINTER(VsxSystemHandle), c_char_p])

            # vsx_SendFirmware()
            cls.send_firmware = cls.__wrap_function("vsx_SendFirmware", c_int32, [POINTER(VsxSystemHandle), c_char_p])

            # vsx_SendXmlDataMessage()
            cls.send_xml_data_message = cls.__wrap_function("vsx_SendXmlDataMessage", c_int32,
                                                            [POINTER(VsxSystemHandle), c_char_p])

            # vsx_SetNetworkSettings() -> change network setting via TCP
            cls.set_network_settings = cls.__wrap_function("vsx_SetNetworkSettings", c_int32,
                                                           [POINTER(VsxSystemHandle), c_char_p, c_char_p, c_char_p])

            # vsx_SetNetworkSettingsViaUdp() -> change network setting via UDP
            cls.set_network_settings_via_udp = cls.__wrap_function("vsx_SetNetworkSettingsViaUdp", c_int32,
                                                                   [c_char_p, c_char_p, c_char_p, c_char_p])

            # vsx_ResetDynamicContainerGrabber()
            cls.reset_dynamic_container_grabber = cls.__wrap_function("vsx_ResetDynamicContainerGrabber", c_int32,
                                                                      [POINTER(VsxSystemHandle), c_int32, c_int32])

            # vsx_GetDataContainer()
            cls.get_data_container = cls.__wrap_function("vsx_GetDataContainer", c_int32,
                                                         [POINTER(VsxSystemHandle),
                                                          POINTER(POINTER(VsxDataContainerHandle)), c_int32])

            # vsx_GetCachedContainer()
            cls.get_cached_container = cls.__wrap_function("vsx_GetCachedContainer", c_int32,
                                                           [POINTER(VsxSystemHandle),
                                                            POINTER(POINTER(VsxDataContainerHandle)), c_int32])

            # vsx_ReleaseDataContainer
            cls.release_data_container = cls.__wrap_function("vsx_ReleaseDataContainer", c_int32,
                                                             [POINTER(POINTER(VsxDataContainerHandle))])

            # vsx_SaveData
            cls.save_data = cls.__wrap_function("vsx_SaveData", c_int32,
                                                [POINTER(VsxDataContainerHandle), c_char_p, c_char_p])

            # vsx_Save3DPointCloudData
            cls.save_3d_point_cloud_data = cls.__wrap_function("vsx_Save3DPointCloudData", c_int32,
                                                               [POINTER(VsxDataContainerHandle), c_char_p, c_char_p,
                                                                c_char_p, c_char_p])

            # vsx_GetCaptureInformation()
            cls.get_capture_information = cls.__wrap_function("vsx_GetCaptureInformation", c_int32,
                                                              [POINTER(VsxDataContainerHandle), c_char_p,
                                                               POINTER(POINTER(VsxCaptureInformationStructure))])

            # vsx_ReleaseCaptureInformation
            cls.release_capture_information = cls.__wrap_function("vsx_ReleaseCaptureInformation", c_int32,
                                                                  [POINTER(POINTER(VsxCaptureInformationStructure))])

            # vsx_GetTransformation()
            cls.get_transformation = cls.__wrap_function("vsx_GetTransformation", c_int32,
                                                         [POINTER(VsxDataContainerHandle), c_char_p,
                                                          POINTER(POINTER(VsxTransformationStructure))])

            # vsx_ReleaseTransformation
            cls.release_transformation = cls.__wrap_function("vsx_ReleaseTransformation", c_int32,
                                                             [POINTER(POINTER(VsxTransformationStructure))])

            # vsx_GetDisparityDescriptor2()
            cls.get_disparity_descriptor2 = cls.__wrap_function("vsx_GetDisparityDescriptor2", c_int32,
                                                                [POINTER(VsxDataContainerHandle), c_char_p,
                                                                 POINTER(POINTER(VsxDisparityDescriptor2Structure))])

            # vsx_ReleaseDisparityDescriptor2
            cls.release_disparity_descriptor2 = cls.__wrap_function("vsx_ReleaseDisparityDescriptor2", c_int32,
                                                                    [POINTER(
                                                                        POINTER(VsxDisparityDescriptor2Structure))])

            # vsx_GetOlr2CaptureInformation()
            cls.get_olr2_capture_information = cls.__wrap_function("vsx_GetOlr2CaptureInformation", c_int32,
                                                                   [POINTER(VsxDataContainerHandle), c_char_p,
                                                                    POINTER(
                                                                        POINTER(VsxOlr2CaptureInformationStructure))])

            # vsx_ReleaseOlr2CaptureInformation
            cls.release_olr2_capture_information = cls.__wrap_function("vsx_ReleaseOlr2CaptureInformation", c_int32,
                                                                       [POINTER(
                                                                           POINTER(
                                                                               VsxOlr2CaptureInformationStructure))])

            # vsx_GetOlr2ModbusData()
            cls.get_olr2_modbus_data = cls.__wrap_function("vsx_GetOlr2ModbusData", c_int32,
                                                           [POINTER(VsxDataContainerHandle), c_char_p,
                                                            POINTER(POINTER(VsxOlr2ModbusDataStructure))])

            # vsx_ReleaseOlr2ModbusData
            cls.release_olr2_modbus_data = cls.__wrap_function("vsx_ReleaseOlr2ModbusData", c_int32,
                                                               [POINTER(
                                                                   POINTER(VsxOlr2ModbusDataStructure))])

            # vsx_GetImage()
            cls.get_image = cls.__wrap_function("vsx_GetImage", c_int32,
                                                [POINTER(VsxDataContainerHandle), c_char_p,
                                                 POINTER(POINTER(VsxImage))])

            # vsx_ReleaseImage
            cls.release_image = cls.__wrap_function("vsx_ReleaseImage", c_int32, [POINTER(POINTER(VsxImage))])

            # vsx_GetLine()
            cls.get_line = cls.__wrap_function("vsx_GetLine", c_int32,
                                               [POINTER(VsxDataContainerHandle), c_char_p,
                                                POINTER(POINTER(VsxLineData))])

            # vsx_ReleaseLine
            cls.release_line = cls.__wrap_function("vsx_ReleaseLine", c_int32, [POINTER(POINTER(VsxLineData))])

            # vsx_GetTagList
            cls.get_tag_list = cls.__wrap_function("vsx_GetTagList", c_int32, [POINTER(VsxDataContainerHandle),
                                                                               POINTER(POINTER(VsxTagList))])

            # vsx_ReleaseTagList
            cls.release_tag_list = cls.__wrap_function("vsx_ReleaseTagList", c_int32, [POINTER(POINTER(VsxTagList))])

            # vsx_GetMissingContainerFramesCounter
            cls.get_missing_container_frames_counter = cls.__wrap_function("vsx_GetMissingContainerFramesCounter",
                                                                           c_int32,
                                                                           [POINTER(VsxSystemHandle), POINTER(c_int32)])

            # vsx_GetDynamicContainerQueueSize
            cls.get_dynamic_container_queue_size = cls.__wrap_function("vsx_GetDynamicContainerQueueSize",
                                                                       c_int32,
                                                                       [POINTER(VsxSystemHandle), POINTER(c_int32)])

            # vsx_GetNumberOfCachedContainers
            cls.get_number_of_cached_containers = cls.__wrap_function("vsx_GetNumberOfCachedContainers",
                                                                      c_int32,
                                                                      [POINTER(VsxSystemHandle), POINTER(c_int32)])

            # vsx_GetDeviceInformation
            cls.get_device_information = cls.__wrap_function("vsx_GetDeviceInformation", c_int32,
                                                             [POINTER(VsxSystemHandle),
                                                              POINTER(POINTER(VsxDevice))])

            # vsx_ReleaseDevice
            cls.release_device = cls.__wrap_function("vsx_ReleaseDevice", c_int32, [POINTER(POINTER(VsxDevice))])

            # vsx_GetUdpDeviceList
            cls.get_udp_device_list = cls.__wrap_function("vsx_GetUdpDeviceList", c_int32,
                                                          [POINTER(POINTER(VsxDeviceList))])

            # vsx_ReleaseDeviceList
            cls.release_device_list = cls.__wrap_function("vsx_ReleaseDeviceList", c_int32,
                                                          [POINTER(POINTER(VsxDeviceList))])

            # vsx_ResetLogMessageGrabber
            cls.reset_log_message_grabber = cls.__wrap_function("vsx_ResetLogMessageGrabber", c_int32,
                                                                [POINTER(VsxSystemHandle), c_int32, c_int32,
                                                                 c_int32])

            # vsx_GetLogMessage()
            cls.get_log_message = cls.__wrap_function("vsx_GetLogMessage", c_int32,
                                                      [POINTER(VsxSystemHandle), POINTER(c_char_p), c_int32])

            # vsx_GetLogMessageQueueSize
            cls.get_log_message_queue_size = cls.__wrap_function("vsx_GetLogMessageQueueSize",
                                                                 c_int32,
                                                                 [POINTER(VsxSystemHandle), POINTER(c_int32)])

            # vsx_GetMissingLogMessagesCounter
            cls.get_missing_log_messages_counter = cls.__wrap_function("vsx_GetMissingLogMessagesCounter",
                                                                       c_int32,
                                                                       [POINTER(VsxSystemHandle), POINTER(c_int32)])

            # vsx_SetSingleParameterValue()
            cls.set_single_parameter_value = cls.__wrap_function("vsx_SetSingleParameterValue", c_int32,
                                                                 [POINTER(VsxSystemHandle), c_int32, c_char_p,
                                                                  c_int32, c_char_p, c_char_p])

            # vsx_SetSingleParameterValueInt32()
            cls.set_single_parameter_value_int32 = cls.__wrap_function("vsx_SetSingleParameterValueInt32", c_int32,
                                                                       [POINTER(VsxSystemHandle), c_int32, c_char_p,
                                                                        c_int32, c_char_p, c_int32])

            # vsx_SetSingleParameterValueDouble()
            cls.set_single_parameter_value_double = cls.__wrap_function("vsx_SetSingleParameterValueDouble", c_int32,
                                                                        [POINTER(VsxSystemHandle), c_int32, c_char_p,
                                                                         c_int32, c_char_p, c_double])

            # vsx_GetSingleParameterValue()
            cls.get_single_parameter_value = cls.__wrap_function("vsx_GetSingleParameterValue", c_int32,
                                                                 [POINTER(VsxSystemHandle), c_int32, c_char_p,
                                                                  c_int32, c_char_p, POINTER(c_char_p)])

            # vsx_GetSingleParameterValueInt32()
            cls.get_single_parameter_value_int32 = cls.__wrap_function("vsx_GetSingleParameterValueInt32", c_int32,
                                                                       [POINTER(VsxSystemHandle), c_int32, c_char_p,
                                                                        c_int32, c_char_p, POINTER(c_int32)])

            # vsx_GetSingleParameterValueDouble()
            cls.get_single_parameter_value_double = cls.__wrap_function("vsx_GetSingleParameterValueDouble", c_int32,
                                                                        [POINTER(VsxSystemHandle), c_int32, c_char_p,
                                                                         c_int32, c_char_p, POINTER(c_double)])

            # vsx_LoadDefaultParameterSetOnDevice
            cls.load_default_parameter_set_on_device = cls.__wrap_function("vsx_LoadDefaultParameterSetOnDevice",
                                                                           c_int32, [POINTER(VsxSystemHandle)])

            # vsx_LoadParameterSetOnDevice
            cls.load_parameter_set_on_device = cls.__wrap_function("vsx_LoadParameterSetOnDevice",
                                                                   c_int32, [POINTER(VsxSystemHandle)])

            # vsx_SaveParameterSetOnDevice
            cls.save_parameter_set_on_device = cls.__wrap_function("vsx_SaveParameterSetOnDevice",
                                                                   c_int32, [POINTER(VsxSystemHandle)])

            # vsx_UploadParameterSet()
            cls.upload_parameter_set = cls.__wrap_function("vsx_UploadParameterSet", c_int32,
                                                           [POINTER(VsxSystemHandle), c_char_p])

            # vsx_DownloadParameterSet()
            cls.download_parameter_set = cls.__wrap_function("vsx_DownloadParameterSet", c_int32,
                                                             [POINTER(VsxSystemHandle), c_char_p])

            # vsx_GetParameterList()
            cls.get_parameter_list = cls.__wrap_function("vsx_GetParameterList", c_int32,
                                                         [POINTER(VsxSystemHandle), POINTER(POINTER(VsxParameterList))])

            # vsx_UploadParameterList()
            cls.upload_parameter_list = cls.__wrap_function("vsx_UploadParameterList", c_int32,
                                                            [POINTER(VsxSystemHandle), POINTER(VsxParameterList)])

            # vsx_SetSingleParameterDouble()
            cls.set_single_parameter_double = cls.__wrap_function("vsx_SetSingleParameterDouble", c_int32,
                                                                  [POINTER(VsxSystemHandle), POINTER(VsxParameter),
                                                                   c_double])

            # vsx_SetSingleParameterInt32()
            cls.set_single_parameter_int32 = cls.__wrap_function("vsx_SetSingleParameterInt32", c_int32,
                                                                 [POINTER(VsxSystemHandle), POINTER(VsxParameter),
                                                                  c_int32])

            # vsx_SetSingleParameterString()
            cls.set_single_parameter_string = cls.__wrap_function("vsx_SetSingleParameterString", c_int32,
                                                                  [POINTER(VsxSystemHandle), POINTER(VsxParameter),
                                                                   c_char_p])

            # vsx_GetSingleParameter()
            cls.get_single_parameter = cls.__wrap_function("vsx_GetSingleParameter", c_int32,
                                                           [POINTER(VsxSystemHandle), POINTER(VsxParameter),
                                                            POINTER(POINTER(VsxParameter))])

            # vsx_ReleaseParameter()
            cls.release_parameter = cls.__wrap_function("vsx_ReleaseParameter", c_int32,
                                                        [POINTER(POINTER(VsxParameter))])

            # vsx_ReleaseParameterList()
            cls.release_parameter_list = cls.__wrap_function("vsx_ReleaseParameterList", c_int32,
                                                             [POINTER(POINTER(VsxParameterList))])

            # vsx_GetResultXml()
            cls.get_result_xml = cls.__wrap_function("vsx_GetResultXml", c_int32,
                                                     [POINTER(VsxDataContainerHandle), c_char_p, POINTER(c_char_p)])

            # vsx_GetResultElementString()
            cls.get_result_element_string = cls.__wrap_function("vsx_GetResultElementString", c_int32,
                                                                [POINTER(VsxDataContainerHandle), c_char_p, c_char_p,
                                                                 POINTER(c_char_p)])

            # vsx_GetResultElementInt32()
            cls.get_result_element_int32 = cls.__wrap_function("vsx_GetResultElementInt32", c_int32,
                                                               [POINTER(VsxDataContainerHandle), c_char_p, c_char_p,
                                                                POINTER(c_int32)])

            # vsx_GetResultElementInt64()
            cls.get_result_element_int64 = cls.__wrap_function("vsx_GetResultElementInt64", c_int32,
                                                               [POINTER(VsxDataContainerHandle), c_char_p, c_char_p,
                                                                POINTER(c_int64)])

            # vsx_GetResultElementDouble()
            cls.get_result_element_double = cls.__wrap_function("vsx_GetResultElementDouble", c_int32,
                                                                [POINTER(VsxDataContainerHandle), c_char_p, c_char_p,
                                                                 POINTER(c_double)])

            # vsx_GetAllDeviceStatusData()
            cls.get_all_device_status_data = cls.__wrap_function("vsx_GetAllDeviceStatusData", c_int32,
                                                                 [POINTER(VsxSystemHandle),
                                                                  POINTER(POINTER(VsxStatusItemList))])

            # vsx_ReleaseStatusItemList()
            cls.release_status_item_list = cls.__wrap_function("vsx_ReleaseStatusItemList", c_int32,
                                                               [POINTER(POINTER(VsxStatusItemList))])

            # vsx_RegisterOnDeviceStatusReceived()
            cls.register_on_device_status_received = cls.__wrap_function("vsx_RegisterOnDeviceStatusReceived", c_int32,
                                                                         [POINTER(VsxSystemHandle),
                                                                          cls.callback_on_device_status_received])

            # vsx_DeregisterOnDeviceStatusReceived()
            cls.deregister_on_device_status_received = cls.__wrap_function("vsx_DeregisterOnDeviceStatusReceived",
                                                                           c_int32, [POINTER(VsxSystemHandle)])

            # vsx_SubscribeToDeviceStatusData()
            cls.subscribe_to_device_status_data = cls.__wrap_function("vsx_SubscribeToDeviceStatusData",
                                                                      c_int32, [POINTER(VsxSystemHandle)])

            # vsx_UnsubscribeToDeviceStatusData()
            cls.unsubscribe_to_device_status_data = cls.__wrap_function("vsx_UnsubscribeToDeviceStatusData",
                                                                        c_int32, [POINTER(VsxSystemHandle)])
