from .Definitions import SerialConnectionType, Strategy, StatusItem, DeviceStatusScope, ValueType, Parameter, \
    DisconnectEvent, SessionTypes
from .DataContainer import DataContainer
from typing import Tuple, Optional, List, Dict, Callable, Union
from deprecation import deprecated
from .Interface import Interface, VsxSystemHandle, VsxDeviceList, VsxDevice, VsxStatusItemList, VsxParameterList, \
    VsxParameter
from ctypes import c_char_p, c_int32, c_double, POINTER, byref


class Sensor(object):
    __status_local_on_connected = {}
    __status_local_on_session_message_received = {}
    __status_local_on_device_status_received = {}

    def __init__(self, *, _is_direct=True):
        self._vsx_handle = None
        self.__status_fn_on_connected = None
        self.__status_fn_on_session_message_received = None
        self.__status_fn_on_device_status_received = None

        # don't initialize me directly
        if _is_direct:
            raise TypeError("create with Sensor.Init*")
        self._vsx_handle = POINTER(VsxSystemHandle)()

    def __del__(self):
        if self._vsx_handle is not None:
            _res = Interface.release_sensor(self._vsx_handle)
            self._vsx_handle = None

    @staticmethod
    def GetLibraryVersion() -> Optional[str]:
        """
        Returns the actual library version.
        """
        Interface.init_driver()
        version_ptr = c_char_p()
        act = Interface.get_library_version(version_ptr)
        if act == 0:
            str_version = version_ptr.value.decode("utf-8")
            _ = Interface.release_string(version_ptr)
            return str_version
        return None

    @staticmethod
    def GetErrorText(error_code: int) -> Optional[str]:
        """
        Return the error text to a given error code.
        It also appends additional text from last error given.
        @param error_code: Input error code
        @return: error text
        """
        Interface.init_driver()
        error_text_ptr = c_char_p()
        act = Interface.get_error_text(error_code, error_text_ptr)
        if act == 0:
            error_text = error_text_ptr.value.decode("utf-8")
            _ = Interface.release_string(error_text_ptr)
            return error_text
        return None

    @property
    def Handle(self) -> Optional[bool]:
        """
        Get the actual used handle from the actual sensor instance
        """
        if self._vsx_handle is None:
            return None
        else:
            return self._vsx_handle.contents.handle

    @classmethod
    def InitTcpSensor(cls, ip_address: str, plugin_name: str):
        """
        Initialize a new tcp based sensor
        @param ip_address: e.g. "192.168.2.4"
        @param plugin_name: Additional functionality for special sensors
        @return: New instance of Sensor class
        """
        Interface.init_driver()
        obj = cls(_is_direct=False)
        if ip_address is None:
            enc_ip_address = c_char_p()
        else:
            enc_ip_address = ip_address.encode("utf-8")

        ret = Interface.init_tcp_sensor(byref(obj._vsx_handle), enc_ip_address, plugin_name.encode("utf-8"))
        if ret != 0:
            raise TypeError(f"Invalid return value {ret}")
        return obj

    @classmethod
    def InitSerialSensor(cls, serial_port: str, baud_rate: int, sensor_type: str, connection_type: SerialConnectionType,
                         plugin_name: str):
        """
        Inits an instance to communicate with a Vsx-Device via serial protocol.
        @param serial_port: The comport of the device
        @param baud_rate: The baudrate of the device
        @param sensor_type: The sensor type of the device
        @param connection_type: The connection type of the device
        @param plugin_name: Additional functionality for special sensors
        @return:
        """
        Interface.init_driver()
        obj = cls(_is_direct=False)

        ret = Interface.init_serial_sensor(byref(obj._vsx_handle), serial_port.encode("utf-8"), baud_rate,
                                           sensor_type.encode("utf-8"), int(connection_type),
                                           plugin_name.encode("utf-8"))
        if ret != 0:
            raise TypeError(f"Invalid return value {ret}")
        return obj

    def TestSystem(self, command: str, input_value: str) -> Tuple[int, Optional[str], int]:
        """
        Sends a test system command to the device
        @param command: The test system command
        @param input_value: Optional input value
        @return: error code, output string of function call, status (1 on success and 0 on failure)
        """
        output_ptr = c_char_p()
        status = c_int32()
        ret = Interface.test_system(self._vsx_handle, command.encode("utf-8"), input_value.encode("utf-8"), output_ptr,
                                    status)
        if ret == 0:
            str_output = output_ptr.value.decode("utf-8")
            ret = Interface.release_string(output_ptr)
            return ret, str_output, status.value
        else:
            return ret, None, 0

    def TestSystemEx(self, command: str, input_value: str, timeout_ms: int) -> Tuple[int, Optional[str], int]:
        """
        Sends a test system command to the device
        @param command: The test system command
        @param input_value: Optional input value
        @param timeout_ms: Timeout for function to execute command (in ms)
        @return: error code, output string of function call, status (1 on success and 0 on failure)
        """
        output_ptr = c_char_p()
        status = c_int32()
        ret = Interface.test_system_ex(self._vsx_handle, command.encode("utf-8"), input_value.encode("utf-8"),
                                       output_ptr,
                                       status, timeout_ms)
        if ret == 0:
            str_output = output_ptr.value.decode("utf-8")
            ret = Interface.release_string(output_ptr)
            return ret, str_output, status.value
        else:
            return ret, None, 0

    def SendFirmware(self, file_name: str) -> int:
        """
        Sends a firmware update file to the device.
        @param file_name: The path and filename of the firmware file
        @return: Returns VSX_STATUS_SUCCESS(0) on success
        """
        return Interface.send_firmware(self._vsx_handle, file_name.encode("utf-8"))

    def UploadData(self, file_name: str) -> int:
        """
        Sends a data file (either image data or dynamic container data) to the device.
        @param file_name: The path and filename of the data file
        @return: Returns VSX_STATUS_SUCCESS(0) on success
        """
        return Interface.upload_data(self._vsx_handle, file_name.encode("utf-8"))

    def UploadParameterSet(self, file_name: str) -> int:
        """
        Uploads a parameter file to the device.
        @param file_name: Path and filename to upload
        @return: Returns VSX_STATUS_SUCCESS(0) on success
        """
        return Interface.upload_parameter_set(self._vsx_handle, file_name.encode("utf-8"))

    def DownloadParameterSet(self, file_name: str) -> int:
        """
        Save the current parameter set to a file.
        @param file_name: Path and file name to save to
        @return: Returns VSX_STATUS_SUCCESS(0) on success
        """
        return Interface.download_parameter_set(self._vsx_handle, file_name.encode("utf-8"))

    def SendXmlDataMessage(self, xml_command: str) -> int:
        """
        Sends a string to the device.
        NOTE: function does not wait for any device reply.
        @param xml_command: Command to send
        @return: Returns VSX_STATUS_SUCCESS(0) on success
        """
        return Interface.send_xml_data_message(self._vsx_handle, xml_command.encode("utf-8"))

    def LoadDefaultParameterSetOnDevice(self) -> int:
        """
        Resets the devices parameters to factory settings and returns a list of the complete
        parameter set of the device including current values.
        @return: Returns VSX_STATUS_SUCCESS(0) on success
        """
        return Interface.load_default_parameter_set_on_device(self._vsx_handle)

    def LoadParameterSetOnDevice(self) -> int:
        """
        Loads the parameter set saved on device and returns a list of the complete parameter
        set of the device including current values.
        @return: Returns VSX_STATUS_SUCCESS(0) on success
        """
        return Interface.load_parameter_set_on_device(self._vsx_handle)

    def SaveParameterSetOnDevice(self) -> int:
        """
        Saves the current parameter set on device. Parameter values will be loaded when
        device starts.
        @return: Returns VSX_STATUS_SUCCESS(0) on success
        """
        return Interface.save_parameter_set_on_device(self._vsx_handle)

    def Connect(self) -> int:
        """
        Connect with the device.
        @return: Returns VSX_STATUS_SUCCESS(0) on success
        """
        return Interface.connect(self._vsx_handle)

    def ConnectEx(self, timeout_ms: int) -> int:
        """
        Connect with the device.
        @param timeout_ms: The timeout for a connection attempt (in ms)
        @return: Returns VSX_STATUS_SUCCESS(0) on success
        """
        return Interface.connect_ex(self._vsx_handle, timeout_ms)

    def ConnectAndLogin(self, username: str, password: str) -> int:
        """
        Connect with the device.
        @param username: username for login
        @param password: password for login
        @return: Returns VSX_STATUS_SUCCESS(0) on success
        """
        return Interface.connect_and_login(self._vsx_handle, username.encode("utf-8"), password.encode("utf-8"))

    def ConnectExAndLogin(self, username: str, password: str, timeout_ms: int) -> int:
        """
        Connect with the device.
        @param username: username for login
        @param password: password for login
        @param timeout_ms: The timeout for a connection attempt (in ms)
        @return: Returns VSX_STATUS_SUCCESS(0) on success
        """
        return Interface.connect_ex_and_login(self._vsx_handle, username.encode("utf-8"), password.encode("utf-8"),
                                              timeout_ms)

    def Login(self, username: str, password: str) -> int:
        """
        Login to the device.
        @param username: username for login
        @param password: password for login
        @return: Returns VSX_STATUS_SUCCESS(0) on success
        """
        return Interface.login(self._vsx_handle, username.encode("utf-8"), password.encode("utf-8"))

    def Logout(self) -> int:
        """
        Logout from device.
        @return: Returns VSX_STATUS_SUCCESS(0) on success
        """
        return Interface.logout(self._vsx_handle)

    def SetPassword(self, authorization_username: str, authorization_password: str, username: str,
                    password: str) -> int:
        """
        Set new password on the device.
        @param authorization_username: username for authorization account
        @param authorization_password: password for authorization account
        @param username: username for account to set new password
        @param password: password for account
        @return: Returns VSX_STATUS_SUCCESS(0) on success
        """
        return Interface.set_password(self._vsx_handle, authorization_username.encode("utf-8"),
                                      authorization_password.encode("utf-8"), username.encode("utf-8"),
                                      password.encode("utf-8"))

    @property
    def Connected(self) -> Optional[bool]:
        """
        Indicates current connection state with the device.
        @return: Returns `True` or `False`
        """
        result = c_int32()
        act = Interface.get_connected(self._vsx_handle, result)
        if act == 0:
            if result.value == 1:
                return True
            else:
                return False
        return None

    def Disconnect(self) -> int:
        """
        Disconnect with the device.
        @return: Returns VSX_STATUS_SUCCESS(0) on success
        """
        return Interface.disconnect(self._vsx_handle)

    @staticmethod
    def _OnDisconnect(handle: c_int32, ip_address: c_char_p, disconnect_event: c_int32, description: c_char_p):
        """
        Callback function for disconnect event (internal usage)
        @param handle: Returns actual used handle
        @param ip_address: Actual used ip address
        @param disconnect_event: Show information about disconnect type
        @param description: Additional description
        """
        if handle in Sensor.__status_local_on_connected:
            func = Sensor.__status_local_on_connected[handle]
            func(handle, ip_address.decode(), DisconnectEvent(disconnect_event), description.decode())

    def RegisterOnDisconnect(self, func: Callable[[int, str, DisconnectEvent, str], None]) -> int:
        """
        Register a callback function for a disconnect event.
        Watch out: The callback will be in another thread (this could be not
        debuggable).

        The function should have the following definition::
            def MyOnDisconnect(handle: int, ip_address: str, disconnectEvent: int, description: str) -> None:
                # your implementation (not working thread, so watch thread safety!)

        @param func: The function should have the following definition
        @return: Return value
        """
        self.__status_fn_on_connected = Interface.callback_on_disconnect(self._OnDisconnect)
        self.__status_local_on_connected[self._vsx_handle.contents.handle] = func
        return Interface.register_on_disconnect(self._vsx_handle, self.__status_fn_on_connected)

    def DeregisterOnDisconnect(self) -> int:
        """
        Function to deregister already existing callback function
        @return: Returns VSX_STATUS_SUCCESS(0) on success
        """
        res = Interface.deregister_on_disconnect(self._vsx_handle)
        self.__status_fn_on_connected = None
        del self.__status_local_on_connected[self._vsx_handle.contents.handle]
        return res

    @staticmethod
    def _OnSessionMessageReceived(handle: c_int32, session_type: c_int32, timeout: c_int32):
        """
        callback for "vsx_OnSessionMessageReceived" (internal usage)
        @param handle: Returns actual used handle
        @param session_type: Session type event
        @param timeout: Timeout, when session will end
        @return: Returns VSX_STATUS_SUCCESS(0) on success
        """
        if handle in Sensor.__status_local_on_connected:
            func = Sensor.__status_local_on_connected[handle]
            func(handle, SessionTypes(session_type), timeout)

    def RegisterOnSessionMessageReceived(self, func: Callable[[int, SessionTypes, int], None]) -> int:
        """
        Register a callback function for a session message event.
        Watch out: The callback will be in another thread (this could be not
        debuggable).

        The function should have the following definition::
            def MyOnSessionMessageReceived(handle: int, session_type: int, timeout: int) -> None:
                # your implementation (not working thread, so watch thread safety!)

        @param func: The function should have the following definition
        @return: Return value
        """
        self.__status_fn_on_session_message_received = Interface.callback_on_session_message_received(
            self._OnSessionMessageReceived)
        self.__status_local_on_session_message_received[self._vsx_handle.contents.handle] = func
        return Interface.register_on_session_message_received(self._vsx_handle,
                                                              self.__status_fn_on_session_message_received)

    def DeregisterOnSessionMessageReceived(self) -> int:
        """
        Function to deregister already existing callback function
        @return: Returns VSX_STATUS_SUCCESS(0) on success
        """
        res = Interface.deregister_on_session_message_received(self._vsx_handle)
        self.__status_fn_on_session_message_received = None
        del self.__status_local_on_session_message_received[self._vsx_handle.contents.handle]
        return res

    def SendSessionKeepAlive(self) -> int:
        """
        Send session keep alive to sensor. Should be the reply from a timeout announcement message
        @return: Returns VSX_STATUS_SUCCESS(0) on success
        """
        return Interface.send_session_keep_alive(self._vsx_handle)

    @deprecated("Use 'ReConnectTcpDevice' instead")
    def ReConnectTcpSensor(self, ip_address: str) -> int:
        return self.ReConnectTcpDevice(ip_address)

    def ReConnectTcpDevice(self, ip_address: str) -> int:
        """
        Disconnects the device and reconnects with new connection settings.
        @param ip_address: The new IPAddress
        @return: Returns VSX_STATUS_SUCCESS(0) on success
        """
        return Interface.reconnect_tcp_device(self._vsx_handle, ip_address.encode("utf-8"))

    def ReConnectAndLoginTcpDevice(self, ip_address: str, username: str, password: str) -> int:
        """
        Disconnects the device and reconnects with new connection settings.
        @param ip_address: The new IPAddress
        @param username: username for login
        @param password: password for login
        @return: Returns VSX_STATUS_SUCCESS(0) on success
        """
        return Interface.reconnect_and_login_tcp_device(self._vsx_handle, ip_address.encode("utf-8"),
                                                        username.encode("utf-8"), password.encode("utf-8"))

    @deprecated("Use 'ReConnectSerialDevice' instead")
    def ReConnectSerialSensor(self, serial_port: str, baud_rate: int, _: str,
                              connection_type: SerialConnectionType) -> int:
        return self.ReConnectSerialDevice(serial_port, baud_rate, connection_type)

    def ReConnectSerialDevice(self, serial_port: str, baud_rate: int,
                              connection_type: SerialConnectionType) -> int:
        """
        Disconnects the device and reconnects with new connection settings.
        @param serial_port: The new serial port
        @param baud_rate: The new baudrate
        @param connection_type: The new connection type
        @return: Returns VSX_STATUS_SUCCESS(0) on success
        """
        return Interface.reconnect_serial_device(self._vsx_handle, serial_port.encode("utf-8"), baud_rate,
                                                 int(connection_type))

    def ReleaseSensor(self) -> int:
        """
        Frees the given sensor
        :return:  Returns VSX_STATUS_SUCCESS(0) on success
        """
        return Interface.release_sensor(byref(self._vsx_handle))

    def SetWaitTimeout(self, timeout_ms) -> int:
        """
        Sets the time in ms, the driver waits for response from device.
        @param timeout_ms: Time in ms
        @return: Returns VSX_STATUS_SUCCESS(0) on success
        """
        return Interface.set_wait_timeout(self._vsx_handle, timeout_ms)

    def GetWaitTimeout(self) -> Tuple[int, int]:
        """
        Gets the time in ms, the driver waits for response from device.
        @return: Returns VSX_STATUS_SUCCESS(0) on success, Time in ms
        """
        timeout_ms = c_int32()
        ret = Interface.get_wait_timeout(self._vsx_handle, timeout_ms)
        return ret, timeout_ms.value

    @deprecated("Use 'SetNetworkSettings' instead")
    def SetNetworkParameter(self, ip_address: str, network_mask: str, gateway: str):
        return self.SetNetworkSettings(ip_address, network_mask, gateway)

    def SetNetworkSettings(self, ip_address: str, network_mask: str, gateway: str):
        """
        Sets the network settings of the device.
        @param ip_address: The new IP Address
        @param network_mask: The new network mask
        @param gateway: The new gateway address
        @return: Returns VSX_STATUS_SUCCESS(0) on success
        """
        return Interface.set_network_settings(self._vsx_handle, ip_address.encode("utf-8"),
                                              network_mask.encode("utf-8"), gateway.encode("utf-8"))

    @staticmethod
    def SetNetworkSettingsViaUdp(mac_address: str, ip_address: str, network_mask: str, gateway: str):
        """
        Sets the network settings of the device identified by the macAddress via UDP.
        @param mac_address: The mac address of the device to set
        @param ip_address: The new IP Address
        @param network_mask: The new network mask
        @param gateway: The new gateway address
        @return: Returns VSX_STATUS_SUCCESS(0) on success
        """
        Interface.init_driver()
        return Interface.set_network_settings_via_udp(mac_address.encode("utf-8"), ip_address.encode("utf-8"),
                                                      network_mask.encode("utf-8"), gateway.encode("utf-8"))

    @deprecated("Use 'SetSingleParameterValue' instead")
    def SetSingleDataValue(self, settings_version: int, configuration_id: str, configuration_version: int,
                           parameter_id: str, value: str) -> int:
        return self.SetSingleParameterValue(settings_version, configuration_id, configuration_version,
                                            parameter_id, value)

    def SetSingleParameterValue(self, settings_version: int, configuration_id: str, configuration_version: int,
                                parameter_id: str, value: Union[str, int, float]) -> int:
        """
        Sets the parameter to a value on the device.
        @param settings_version: The settings version of the parameter which should be set
        @param configuration_id: The config id of the parameter which should be set
        @param configuration_version: The config version of the parameter which should be se
        @param parameter_id: The id of the parameter which should be set
        @param value: Value as string, float or integer number
        @return: Returns VSX_STATUS_SUCCESS(0) on success
        """
        if isinstance(value, int):
            return Interface.set_single_parameter_value_int32(self._vsx_handle, settings_version,
                                                              configuration_id.encode("utf-8"),
                                                              configuration_version, parameter_id.encode("utf-8"),
                                                              value)
        elif isinstance(value, float):
            return Interface.set_single_parameter_value_double(self._vsx_handle, settings_version,
                                                               configuration_id.encode("utf-8"),
                                                               configuration_version, parameter_id.encode("utf-8"),
                                                               value)
        elif isinstance(value, str):
            return Interface.set_single_parameter_value(self._vsx_handle, settings_version,
                                                        configuration_id.encode("utf-8"),
                                                        configuration_version, parameter_id.encode("utf-8"),
                                                        value.encode("utf-8"))
        else:
            raise TypeError("Unknown valueType ")

    @deprecated("Use 'GetSingleParameterValue' instead")
    def GetSingleDataValue(self, settings_version: int, configuration_id: str, configuration_version: int,
                           parameter_id: str) -> Tuple[int, str]:
        return self.GetSingleParameterValue(settings_version, configuration_id, configuration_version, parameter_id)

    def GetSingleParameterValue(self, settings_version: int, configuration_id: str, configuration_version: int,
                                parameter_id: str) -> Tuple[int, Optional[str]]:
        """
        Returns the current value of the given parameter from device.
        @param settings_version: The settings version of the parameter its value is asked for
        @param configuration_id: The config id of the parameter its value is asked for
        @param configuration_version: The config version of the parameter its value is asked for
        @param parameter_id: The id of the parameter its value is asked for
        @return: Returns VSX_STATUS_SUCCESS(0) on success and value in string representation
        """
        value_ptr = c_char_p()
        ret = Interface.get_single_parameter_value(self._vsx_handle, settings_version,
                                                   configuration_id.encode("utf-8"),
                                                   configuration_version, parameter_id.encode("utf-8"),
                                                   byref(value_ptr))
        if ret == 0:
            value = value_ptr.value.decode("utf-8")
            ret = Interface.release_string(value_ptr)
            return ret, value
        else:
            return ret, None

    def GetSingleParameterValueInt32(self, settings_version: int, configuration_id: str, configuration_version: int,
                                     parameter_id: str) -> Tuple[int, Optional[int]]:
        """
        Returns the current value of the given parameter from device.
        @param settings_version: The settings version of the parameter its value is asked for
        @param configuration_id: The config id of the parameter its value is asked for
        @param configuration_version: The config version of the parameter its value is asked for
        @param parameter_id: The id of the parameter its value is asked for
        @return: Returns VSX_STATUS_SUCCESS(0) on success and value in integer representation
        """
        value_ptr = c_int32()
        ret = Interface.get_single_parameter_value_int32(self._vsx_handle, settings_version,
                                                         configuration_id.encode("utf-8"),
                                                         configuration_version, parameter_id.encode("utf-8"),
                                                         value_ptr)
        if ret == 0:
            return ret, value_ptr.value
        else:
            return ret, None

    def GetSingleParameterValueDouble(self, settings_version: int, configuration_id: str, configuration_version: int,
                                      parameter_id: str) -> Tuple[int, Optional[float]]:
        """
        Returns the current value of the given parameter from device.
        @param settings_version: The settings version of the parameter its value is asked for
        @param configuration_id: The config id of the parameter its value is asked for
        @param configuration_version: The config version of the parameter its value is asked for
        @param parameter_id: The id of the parameter its value is asked for
        @return: Returns VSX_STATUS_SUCCESS(0) on success and value in float representation
        """
        value_ptr = c_double()
        ret = Interface.get_single_parameter_value_double(self._vsx_handle, settings_version,
                                                          configuration_id.encode("utf-8"),
                                                          configuration_version, parameter_id.encode("utf-8"),
                                                          value_ptr)
        if ret == 0:
            return ret, value_ptr.value
        else:
            return ret, None

    def GetParameterList(self) -> Tuple[int, Optional[List[Parameter]]]:
        """
        Returns a list of the complete parameter set of the device including their current
        values. The list shows the current state of the parameters.
        @return: Returns VSX_STATUS_SUCCESS(0) on success and parameter list data
        """
        result_ptr = POINTER(VsxParameterList)()
        ret = Interface.get_parameter_list(self._vsx_handle, byref(result_ptr))
        if ret == 0:
            res = Sensor.__generate_parameter_list_from_vsx(result_ptr)

            ret = Interface.release_parameter_list(byref(result_ptr))
            return ret, res
        else:
            return ret, None

    def UploadParameterList(self, parameter_list_data: List[Parameter]) -> int:
        """
        Uploads a parameter list to the device.
        @param parameter_list_data:
        @return: Returns VSX_STATUS_SUCCESS(0) on success
        """
        result = Sensor.__generate_parameter_list(parameter_list_data)
        ret = Interface.upload_parameter_list(self._vsx_handle, byref(result))
        return ret

    def SetSingleParameter(self, parameter: Parameter) -> int:
        """
        Sets the parameter to a value on the device.
        @param parameter: The parameter the value should be set from
        @return: Returns VSX_STATUS_SUCCESS(0) on success
        """
        result = Sensor.__generate_parameter(parameter)
        if parameter.valueType == ValueType.DOUBLE or parameter.valueType == ValueType.FLOAT:
            ret = Interface.set_single_parameter_double(self._vsx_handle, byref(result), result.valueDouble)
        elif parameter.valueType == ValueType.INT or parameter.valueType == ValueType.BOOL or \
                parameter.valueType == ValueType.INT16 or parameter.valueType == ValueType.UINT:
            ret = Interface.set_single_parameter_int32(self._vsx_handle, byref(result), result.valueInt)
        elif parameter.valueType == ValueType.STRING or parameter.valueType == ValueType.ENUM:
            ret = Interface.set_single_parameter_string(self._vsx_handle, byref(result), result.valueString)
        else:
            raise TypeError("Unknown valueType ")
        return ret

    def GetSingleParameter(self, parameter: Parameter) -> Tuple[int, Optional[Parameter]]:
        """
        Returns the current value of the given parameter from device.
        @param parameter: The parameter its value is asked for
        @return: Returns VSX_STATUS_SUCCESS(0) on success and the new parameter
        """
        parameter_in = Sensor.__generate_parameter(parameter)
        parameter_out_ptr = POINTER(VsxParameter)()
        ret = Interface.get_single_parameter(self._vsx_handle, byref(parameter_in), byref(parameter_out_ptr))
        if ret == 0:
            res = Sensor.__generate_parameter_from_vsx(parameter_out_ptr.contents)
            ret = Interface.release_parameter(byref(parameter_out_ptr))
            if res.name is None:
                res.name = parameter.name
            if len(res.enumItems) == 0:
                res.enumItems = parameter.enumItems
            return ret, res
        else:
            return ret, None

    @staticmethod
    def __generate_parameter(parameter: Parameter) -> VsxParameter:
        vsx_parameter = VsxParameter()
        vsx_parameter.settingsVersion = parameter.settingsVersion
        vsx_parameter.configVersion = parameter.configVersion
        vsx_parameter.configId = parameter.configId.encode()
        vsx_parameter.parameterId = parameter.parameterId.encode()
        vsx_parameter.name = parameter.name.encode()
        vsx_parameter.valueType = parameter.valueType

        if parameter.valueType == ValueType.DOUBLE or parameter.valueType == ValueType.FLOAT:
            vsx_parameter.valueDouble = parameter.value
        elif parameter.valueType == ValueType.INT or \
                parameter.valueType == ValueType.INT16 or parameter.valueType == ValueType.UINT:
            vsx_parameter.valueInt = parameter.value
        elif parameter.valueType == ValueType.BOOL:
            vsx_parameter.valueInt = 1 if parameter.value is True else False
        elif parameter.valueType == ValueType.STRING or parameter.valueType == ValueType.ENUM:
            vsx_parameter.valueString = str(parameter.value).encode()
        else:
            raise TypeError("Unknown valueType ")
        return vsx_parameter

    @staticmethod
    def __generate_parameter_list(parameter_list: List[Parameter]):
        length = len(parameter_list)
        vsx_parameter_list = VsxParameterList(length)
        for i in range(len(parameter_list)):
            parameter = Sensor.__generate_parameter(parameter_list[i])
            vsx_parameter_list.parameters[i] = parameter
        return vsx_parameter_list

    @staticmethod
    def __generate_parameter_from_vsx(parameter: VsxParameter) -> Parameter:
        if parameter.name is not None:
            name = parameter.name.decode("utf-8")
        else:
            name = None
        parameter_id = parameter.parameterId.decode("utf-8")
        config_id = parameter.configId.decode("utf-8")
        config_version = parameter.configVersion
        settings_version = parameter.settingsVersion
        value_type = ValueType(parameter.valueType)
        enum_items = {}

        if value_type == ValueType.DOUBLE or value_type == ValueType.FLOAT:
            value = parameter.valueDouble
        elif value_type == ValueType.INT or value_type == ValueType.INT16 or value_type == ValueType.UINT:
            value = parameter.valueInt
        elif value_type == ValueType.BOOL:
            value = True if parameter.valueInt == 1 else False
        elif value_type == ValueType.STRING:
            value = parameter.valueString.decode("utf-8")
        elif value_type == ValueType.ENUM:
            value = parameter.valueString.decode("utf-8")
            enum_items = {parameter.enumItemList[i].id.decode("utf-8"): parameter.enumItemList[i].name.decode("utf-8")
                          for i in range(parameter.enumItemListLength)}
        else:
            value = None

        parameter = Parameter(settings_version, config_id, config_version, parameter_id, name, value, value_type,
                              enum_items)
        return parameter

    @staticmethod
    def __generate_parameter_list_from_vsx(parameter_list_ptr: POINTER(VsxParameterList)):
        parameter_list = []
        for i in range(parameter_list_ptr.contents.length):
            parameter = Sensor.__generate_parameter_from_vsx(parameter_list_ptr.contents.parameters[i])
            parameter_list.append(parameter)
        return parameter_list

    def ResetLogMessageGrabber(self, buffer_size: int, type_mask: int, strategy: Strategy) -> int:
        """
        Starts the internal log message grabber.
        @param buffer_size: The maximum number of items which will be internally saved, if less than 0, number is infinity.
        @param type_mask: Mask which log message types will be send by device.
        @param strategy: The strategy, which items will be discarded if maximum number of items is reached
        @return: Returns VSX_STATUS_SUCCESS(0) on success
        """
        return Interface.reset_log_message_grabber(self._vsx_handle, buffer_size, type_mask, strategy)

    def GetLogMessage(self, timeout_ms: int) -> Tuple[int, Optional[str]]:
        """
        Gets the oldest saved item and removes it internally
        @param timeout_ms: The maximum time in ms to try reading an item
        @return: Returns VSX_STATUS_SUCCESS(0) on success
        """
        log_ptr = c_char_p()
        ret = Interface.get_log_message(self._vsx_handle, log_ptr, timeout_ms)
        if ret == 0:
            log = log_ptr.value.decode("utf-8")
            ret = Interface.release_string(log_ptr)
            return ret, log
        else:
            return ret, None

    @property
    def LogMessageQueueSize(self) -> Optional[int]:
        """
        Gets the current size of the log message queue
        @return: Actual log message counter
        """
        result = c_int32()
        act = Interface.get_log_message_queue_size(self._vsx_handle, result)
        if act == 0:
            return result.value
        return None

    @property
    def MissingLogMessagesCounter(self) -> Optional[int]:
        """
        Gets the missing log messages counter for log message grabbing
        @return: Missing log messages counter
        """
        result = c_int32()
        act = Interface.get_missing_log_messages_counter(self._vsx_handle, result)
        if act == 0:
            return result.value
        return None

    @deprecated("Use 'ResetDynamicContainerGrabberEx' without 'frame_counter' instead")
    def ResetDynamicContainerGrabber(self, buffer_size: int, _: int, strategy: Strategy) -> int:
        return self.ResetDynamicContainerGrabberEx(buffer_size, strategy)

    def ResetDynamicContainerGrabberEx(self, buffer_size: int, strategy: Strategy) -> int:
        """
        Restarts the internal dynamic container grabber.
        @param buffer_size: The maximum number of items which will be internally saved, if less than 0, number is infinity.
        @param strategy: The strategy, which items will be discarded if maximum number of items is reached.
        @return: Returns VSX_STATUS_SUCCESS(0) on success
        """
        return Interface.reset_dynamic_container_grabber(self._vsx_handle, buffer_size, int(strategy))

    def GetDataContainer(self, timeout_ms: int) -> Tuple[int, Optional[DataContainer]]:
        """
        Gets the oldest saved item and removes it internally.
        @param timeout_ms: The maximum time in ms to try reading an item.
        @return: Returns VSX_STATUS_SUCCESS(0) on success and new dynamic container
        """
        obj = DataContainer(_is_direct=False)

        ret = Interface.get_data_container(self._vsx_handle, byref(obj.data_container_handle), timeout_ms)
        if ret != 0:
            return ret, None
        return ret, obj

    def GetCachedContainer(self, position: int) -> Tuple[int, Optional[DataContainer]]:
        """
        Gets a cached dynamic container.
        @param position: Position of the container in cache
        @return: Returns VSX_STATUS_SUCCESS(0) on success and new dynamic container
        """
        obj = DataContainer(_is_direct=False)

        ret = Interface.get_cached_container(self._vsx_handle, byref(obj.data_container_handle), position)
        if ret != 0:
            return ret, None
        return ret, obj

    @property
    def MissingContainerFramesCounter(self) -> Optional[int]:
        """
        Gets the missing frame counter from image grabbing
        @return: Actual missing frame counter
        """
        result = c_int32()
        act = Interface.get_missing_container_frames_counter(self._vsx_handle, result)
        if act == 0:
            return result.value
        return None

    @property
    def DynamicContainerQueueSize(self) -> Optional[int]:
        """
        Gets the current size of the dynamic container message queue
        @return: Returns actual queue size
        """
        result = c_int32()
        act = Interface.get_dynamic_container_queue_size(self._vsx_handle, result)
        if act == 0:
            return result.value
        return None

    @property
    def NumberOfCachedContainers(self) -> Optional[int]:
        """
        Gets the current number of cached container messages
        @return: Returns actual cached containers
        """
        result = c_int32()
        act = Interface.get_number_of_cached_containers(self._vsx_handle, result)
        if act == 0:
            return result.value
        return None

    @deprecated("Use 'GetDeviceInformation' instead")
    def GetCurrentDeviceInformation(self) -> Tuple[int, Optional[Dict[str, str]]]:
        return self.GetDeviceInformation()

    def GetDeviceInformation(self) -> Tuple[int, Optional[Dict[str, str]]]:
        """
        Returns a device object with network information about the current device.
        @return: Returns VSX_STATUS_SUCCESS(0) on success and device data object
        """
        result_ptr = POINTER(VsxDevice)()
        ret = Interface.get_device_information(self._vsx_handle, byref(result_ptr))

        if ret == 0:
            results = Sensor.__get_results_from_device(result_ptr.contents)
            ret = Interface.release_device(byref(result_ptr))
            return ret, results
        else:
            return ret, None

    @staticmethod
    def __get_results_from_device(device: VsxDevice) -> Dict[str, str]:
        results = {}
        if device.ipAddress:
            results['ipAddress'] = device.ipAddress.decode("utf-8")
        if device.networkMask:
            results['networkMask'] = device.networkMask.decode("utf-8")
        if device.gateway:
            results['gateway'] = device.gateway.decode("utf-8")
        if device.macAddress:
            results['macAddress'] = device.macAddress.decode("utf-8")
        if device.firmwareVersion:
            results['firmwareVersion'] = device.firmwareVersion.decode("utf-8")
        if device.sensorType:
            results['sensorType'] = device.sensorType.decode("utf-8")
        if device.sensorName:
            results['sensorName'] = device.sensorName.decode("utf-8")
        results['busy'] = bool(device.busy)
        results['deviceVsxVersionMajor'] = device.deviceVsxVersionMajor
        results['deviceVsxVersionMinor'] = device.deviceVsxVersionMinor
        if device.comPort:
            results['comPort'] = device.comPort.decode("utf-8")
        results['baudrate'] = device.baudrate
        if device.headAddress:
            results['headAddress'] = device.headAddress.decode("utf-8")
        results['isLoginNeeded'] = bool(device.isLoginNeeded)
        return results

    @staticmethod
    def GetUdpDeviceList() -> Tuple[int, Optional[List[Dict[str, str]]]]:
        """
        Searches for all devices in a subnet via udp and returns a list with all devices found.
        @return: Returns VSX_STATUS_SUCCESS(0) on success and list of device information
        """
        Interface.init_driver()
        result_ptr = POINTER(VsxDeviceList)()
        ret = Interface.get_udp_device_list(byref(result_ptr))
        if ret == 0:
            results_list = []
            for i in range(result_ptr.contents.length):
                results = Sensor.__get_results_from_device(result_ptr.contents.devices[i])
                results_list.append(results)

            ret = Interface.release_device_list(byref(result_ptr))
            return ret, results_list
        else:
            return ret, None

    def GetAllDeviceStatusData(self) -> Tuple[int, Optional[List[StatusItem]]]:
        """
        Get the full status data set from device.
        @return: Returns VSX_STATUS_SUCCESS(0) on success and list of status data objects
        """
        result_ptr = POINTER(VsxStatusItemList)()
        ret = Interface.get_all_device_status_data(self._vsx_handle, byref(result_ptr))
        if ret == 0:
            res = Sensor.__generate_status_data_list(result_ptr)

            ret = Interface.release_status_item_list(byref(result_ptr))
            return ret, res
        else:
            return ret, None

    @staticmethod
    def __generate_status_data_list(status_item_list_ptr: POINTER(VsxStatusItemList)):
        status_item_list = []
        for i in range(status_item_list_ptr.contents.length):
            name = status_item_list_ptr.contents.statusItems[i].name.decode("utf-8")
            status_item_id = status_item_list_ptr.contents.statusItems[i].statusItemId.decode("utf-8")
            configuration_class = status_item_list_ptr.contents.statusItems[i].configurationClass.decode("utf-8")
            config_version = status_item_list_ptr.contents.statusItems[i].configVersion
            settings_version = status_item_list_ptr.contents.statusItems[i].settingsVersion
            value_type = ValueType(status_item_list_ptr.contents.statusItems[i].valueType)
            if value_type == ValueType.DOUBLE or value_type == ValueType.FLOAT:
                value = status_item_list_ptr.contents.statusItems[i].valueDouble
            elif value_type == ValueType.INT or value_type == ValueType.INT16:
                value = status_item_list_ptr.contents.statusItems[i].valueInt
            elif value_type == ValueType.BOOL:
                value = True if status_item_list_ptr.contents.statusItems[i].valueInt == 1 else False
            elif value_type == ValueType.STRING or value_type == ValueType.ENUM:
                value = status_item_list_ptr.contents.statusItems[i].valueString.decode("utf-8")
            else:
                value = None

            time = status_item_list_ptr.contents.statusItems[i].time
            sensor_time = status_item_list_ptr.contents.statusItems[i].sensorTime

            status_item = StatusItem(settings_version, configuration_class, config_version, status_item_id, name, value,
                                     value_type, time, sensor_time)
            status_item_list.append(status_item)
        return status_item_list

    @staticmethod
    def _OnDeviceStatusReceived(handle: c_int32, device_status_scope: c_int32,
                                status_item_list: POINTER(VsxStatusItemList)):
        """
        callback for "vsx_OnDeviceStatusReceived" (internal)
        """
        if handle in Sensor.__status_local_on_device_status_received:
            res = Sensor.__generate_status_data_list(status_item_list)
            func = Sensor.__status_local_on_device_status_received[handle]
            func(handle, DeviceStatusScope(device_status_scope), res)

    def RegisterOnDeviceStatusReceived(self, func: Callable[[int, DeviceStatusScope, List[StatusItem]], None]) -> int:
        """
        Register a callback function for a disconnect event.
        Watch out: The callback will be in another thread (this could be not
        debuggable).

        The function should have the following definition::
            def MyOnDeviceStatusReceived(handle: int, device_status_scope: DeviceStatusScope,
                    status_item_list: List[StatusItem]) -> None:
                # your implementation (not working thread, so watch thread safety!)

        @param func: The function should have the following definition:
        @return:
        """
        self.__status_fn_on_device_status_received = Interface.callback_on_device_status_received(
            self._OnDeviceStatusReceived)
        self.__status_local_on_device_status_received[self._vsx_handle.contents.handle] = func
        return Interface.register_on_device_status_received(self._vsx_handle,
                                                            self.__status_fn_on_device_status_received)

    def DeregisterOnDeviceStatusReceived(self) -> int:
        """
        Function to deregister already existing callback function
        @return: Returns VSX_STATUS_SUCCESS(0) on success
        """
        res = Interface.deregister_on_device_status_received(self._vsx_handle)
        self.__status_fn_on_device_status_received = None
        del self.__status_local_on_device_status_received[self._vsx_handle.contents.handle]
        return res

    def SubscribeToDeviceStatusData(self) -> int:
        """
        Subscribe status data from sensor to the client.
        This will send periodically or in case of a problem status data to the client.
        This need a registered callback for "vsx_OnDeviceStatusReceived".
        @return: Returns VSX_STATUS_SUCCESS(0) on success
        """
        return Interface.subscribe_to_device_status_data(self._vsx_handle)

    def UnsubscribeToDeviceStatusData(self) -> int:
        """
        Unsubscribe status data from sensor.
        @return: Returns VSX_STATUS_SUCCESS(0) on success
        """
        return Interface.unsubscribe_to_device_status_data(self._vsx_handle)
