"""
@file
@brief                  
            

@copyright  2022 Pepperl + Fuchs SE, Mannheim
            Proprietary material, all rights reserved, use or disclosure forbidden.
"""
import time
from typing import List

from VsxProtocolDriver import Sensor, StatusItem, DeviceStatusScope, DisconnectEvent

import VsxProtocolDriver
import re
import pytest


def test_get_error_text():
    ret = VsxProtocolDriver.Sensor.GetErrorText(-0x8001)
    assert ret == 'VSX_STATUS_ERROR_UNABLE_TO_ALLOCATE_VSX_SYSTEM'
    ret = VsxProtocolDriver.Sensor.GetErrorText(0)
    assert ret == 'VSX_STATUS_SUCCESS'
    ret = VsxProtocolDriver.Sensor.GetErrorText(123456)
    assert ret == 'Unknown Error Code 123456.'


def test_get_library_version(vsx_sensor):
    # Official semver regex
    # https://semver.org/#is-there-a-suggested-regular-expression-regex-to-check-a-semver-string
    regex = r"^(?P<major>0|[1-9]\d*)\.(?P<minor>0|[1-9]\d*)\.(?P<patch>0|[1-9]\d*)"
    r"(?:-(?P<prerelease>(?:0|[1-9]\d*|\d*[a-zA-Z-][0-9a-zA-Z-]*)(?:\.(?:0|[1-9]"
    r"\d*|\d*[a-zA-Z-][0-9a-zA-Z-]*))*))?(?:\+(?P<buildmetadata>[0-9a-zA-Z-]+(?:\.[0-9a-zA-Z-]+)*))?$"

    ret = VsxProtocolDriver.Sensor.GetLibraryVersion()
    assert re.match(regex, ret)

    ret = vsx_sensor.GetLibraryVersion()
    assert re.match(regex, ret)


def test_get_udp_device_list(vsx_sensor):
    ret, results_list = VsxProtocolDriver.Sensor.GetUdpDeviceList()
    assert ret == 0


def test_get_current_device_information(vsx_sensor):
    res = vsx_sensor.Connect()
    assert res == 0

    ret, results_list = vsx_sensor.GetCurrentDeviceInformation()
    assert ret == 0


def test_get_parameter(vsx_sensor):
    res = vsx_sensor.Connect()
    assert res == 0

    ret = vsx_sensor.SetSingleParameterValue(1, 'Base', 1, 'TargetPositionDeg', 1.5)
    assert ret == 0

    ret = vsx_sensor.SetSingleParameterValue(1, 'Base', 1, 'TargetPositionDeg', 15)
    assert ret == 0

    ret = vsx_sensor.SetSingleParameterValue(1, 'Base', 1, 'ExposureTime', 123)
    assert ret == 0

    ret, val = vsx_sensor.GetSingleParameterValue(1, 'Base', 1, 'ExposureTime')
    assert ret == 0

    ret, val = vsx_sensor.GetSingleParameterValueInt32(1, 'Base', 1, 'ExposureTime')
    assert ret == 0

    ret, val = vsx_sensor.GetSingleParameterValueDouble(1, 'Base', 1, 'ExposureTime')
    assert ret == 0

    ret, val = vsx_sensor.GetSingleParameterValueDouble(1, 'Base', 1, 'TriggerSource')
    assert ret == 0


def test_wait_timeout(vsx_sensor):
    res, timeout = vsx_sensor.GetWaitTimeout()
    assert res == 0

    res = vsx_sensor.SetWaitTimeout(timeout + 1)
    assert res == 0

    res, timeout1 = vsx_sensor.GetWaitTimeout()
    assert res == 0
    assert timeout1 == timeout + 1

    res = vsx_sensor.SetWaitTimeout(timeout)
    assert res == 0


def MyOnDisconnect(handle: int, ip_address: str, event: DisconnectEvent, description: str) -> None:
    with open("c:\\temp\\sample.txt", "a") as file_object:
        file_object.write(f"H: {handle} I: {ip_address} D: {description} \n")
    print(f"H: {handle} I: {ip_address} D: {description} E: {event.name}", flush=True)


def test_is_connected(vsx_sensor):
    res = vsx_sensor.Connect()
    assert res == 0

    time.sleep(1)
    print("Register")

    assert vsx_sensor.Connected

    vsx_sensor.RegisterOnDisconnect(MyOnDisconnect)

    res = vsx_sensor.Disconnect()
    time.sleep(1)
    vsx_sensor.DeregisterOnDisconnect()

    assert res == 0

    assert not vsx_sensor.Connected


def test_direct_instance():
    with pytest.raises(TypeError):
        vsx = VsxProtocolDriver.Sensor()


@pytest.fixture
def vsx_sensor():
    vsx = VsxProtocolDriver.Sensor.InitTcpSensor("192.168.2.4", "SR3D_STEREO")
    yield vsx
    ret = vsx.Disconnect()
    del vsx


def test_get_parameter_list(vsx_sensor):
    res = vsx_sensor.Connect()
    assert res == 0

    ret, res = vsx_sensor.GetParameterList()
    assert ret == 0

    ret = vsx_sensor.SetSingleParameter(res[1])
    assert ret == 0

    ret, res2 = vsx_sensor.GetSingleParameter(res[1])
    assert ret == 0

    # Needs VSX Version >= 4
    ret = vsx_sensor.UploadParameterList(res)
    assert ret == 0 or ret == -11


def test_set_parameter_double(vsx_sensor):
    res = vsx_sensor.Connect()
    assert res == 0

    ret, res = vsx_sensor.GetParameterList()
    assert ret == 0

    value = res[3].value
    res[3].value = 5.0

    ret = vsx_sensor.SetSingleParameter(res[3])
    assert ret == 0

    ret, res2 = vsx_sensor.GetSingleParameter(res[3])
    assert ret == 0
    assert res2.value == 5.0

    res[3].value = value
    ret = vsx_sensor.SetSingleParameter(res[3])
    assert ret == 0

    ret, res2 = vsx_sensor.GetSingleParameter(res[3])
    assert ret == 0
    assert res2.value == value


def test_set_parameter_int32(vsx_sensor):
    res = vsx_sensor.Connect()
    assert res == 0

    ret, res = vsx_sensor.GetParameterList()
    assert ret == 0

    value = res[4].value
    res[4].value = 4321

    ret = vsx_sensor.SetSingleParameter(res[4])
    assert ret == 0

    ret, res2 = vsx_sensor.GetSingleParameter(res[4])
    assert ret == 0
    assert res2.value == 4321

    res[4].value = value
    ret = vsx_sensor.SetSingleParameter(res[4])
    assert ret == 0

    ret, res2 = vsx_sensor.GetSingleParameter(res[4])
    assert ret == 0
    assert res2.value == value


def test_set_parameter_string(vsx_sensor):
    res = vsx_sensor.Connect()
    assert res == 0

    ret, res = vsx_sensor.GetParameterList()
    assert ret == 0

    value = res[2].value
    res[2].value = "FallingEdge"

    ret = vsx_sensor.SetSingleParameter(res[2])
    assert ret == 0

    ret, res2 = vsx_sensor.GetSingleParameter(res[2])
    assert ret == 0
    assert res2.value == "FallingEdge"

    res[2].value = value
    ret = vsx_sensor.SetSingleParameter(res[2])
    assert ret == 0

    ret, res2 = vsx_sensor.GetSingleParameter(res[2])
    assert ret == 0
    assert res2.value == value


def test_get_all_status_data(vsx_sensor):
    res = vsx_sensor.Connect()
    assert res == 0

    ret, res = vsx_sensor.GetAllDeviceStatusData()
    assert ret == 0


def test_disconnect_problem(vsx_sensor):
    res = vsx_sensor.Connect()
    assert res == 0
    res2 = vsx_sensor.TestSystem("VersionId", "CPU")
    assert res2[0] == 0
    res = vsx_sensor.Disconnect()
    assert res == 0
    pass


def test_for_bug_in_2_5_1(vsx_sensor):
    res = vsx_sensor.Connect()
    assert res == 0

    res = vsx_sensor.SetSingleDataValue(1, "Command", 1, "TriggerSoftware", "1")

    assert res == 0, vsx_sensor.GetErrorText(res)

    res, val = vsx_sensor.GetSingleDataValue(1, "Base", 1, "OutputMode")
    assert res == 0


@pytest.mark.xfail
def test_get_line(vsx_sensor):
    res = vsx_sensor.Connect()
    assert res == 0

    res = vsx_sensor.SetSingleDataValue(1, "Base", 1, "TriggerSource", "Software")
    assert res == 0
    res = vsx_sensor.SetSingleDataValue(1, "Base", 1, "OutputMode", "DisparityCAndRaw")
    assert res == 0
    res = vsx_sensor.SetSingleDataValue(1, "Command", 1, "TriggerSoftware", "1")
    assert res == 0
    res = vsx_sensor.SetSingleDataValue(1, "Command", 1, "TriggerSoftware", "1")
    assert res == 0

    res, dch = vsx_sensor.GetDataContainer(10000);
    assert res == 0, vsx_sensor.GetErrorText(res)

    res, lines, result = dch.GetLine("Line")
    assert res == 0


def test_get_capture_information(vsx_sensor):
    res = vsx_sensor.Connect()
    assert res == 0

    res = vsx_sensor.SetSingleDataValue(1, "Base", 1, "TriggerSource", "Software")
    assert res == 0
    res = vsx_sensor.SetSingleDataValue(1, "Base", 1, "OutputMode", "DisparityCAndRaw")
    assert res == 0
    res = vsx_sensor.SetSingleDataValue(1, "Command", 1, "TriggerSoftware", "1")
    assert res == 0
    res = vsx_sensor.SetSingleDataValue(1, "Command", 1, "TriggerSoftware", "1")
    assert res == 0

    res, dch = vsx_sensor.GetDataContainer(10000);
    assert res == 0, vsx_sensor.GetErrorText(res)

    res, info = dch.GetCaptureInformation("CaptureInfo")
    assert res == 0
    assert info.exposure_time > 0


def test_get_transformation(vsx_sensor):
    res = vsx_sensor.Connect()
    assert res == 0

    res = vsx_sensor.SetSingleDataValue(1, "Base", 1, "TriggerSource", "Software")
    assert res == 0
    res = vsx_sensor.SetSingleDataValue(1, "Base", 1, "OutputMode", "DisparityCAndRaw")
    assert res == 0
    res = vsx_sensor.SetSingleDataValue(1, "Command", 1, "TriggerSoftware", "1")
    assert res == 0
    res = vsx_sensor.SetSingleDataValue(1, "Command", 1, "TriggerSoftware", "1")
    assert res == 0

    res, dch = vsx_sensor.GetDataContainer(10000);
    assert res == 0, vsx_sensor.GetErrorText(res)

    res, transformation = dch.GetTransformation("Transformation")
    assert res == 0
    assert transformation.translation_tx != float('NaN')


def test_get_disparity_descriptor2(vsx_sensor):
    res = vsx_sensor.Connect()
    assert res == 0

    res = vsx_sensor.SetSingleDataValue(1, "Base", 1, "TriggerSource", "Software")
    assert res == 0
    res = vsx_sensor.SetSingleDataValue(1, "Base", 1, "OutputMode", "DisparityCAndRaw")
    assert res == 0
    res = vsx_sensor.SetSingleDataValue(1, "Command", 1, "TriggerSoftware", "1")
    assert res == 0
    res = vsx_sensor.SetSingleDataValue(1, "Command", 1, "TriggerSoftware", "1")
    assert res == 0

    res, dch = vsx_sensor.GetDataContainer(10000);
    assert res == 0, vsx_sensor.GetErrorText(res)

    res, disp = dch.GetDisparityDescriptor2("DisparityDesc")
    assert res == 0
    assert disp.focal_length != float('NaN')


global_status_item_list = None
global_handle = None
global_device_status_scope = None


def MyOnDeviceStatusReceived(handle: int, device_status_scope: DeviceStatusScope,
                             status_item_list: List[StatusItem]) -> None:
    global global_status_item_list
    global global_handle
    global global_device_status_scope
    global_status_item_list = status_item_list
    global_handle = handle
    global_device_status_scope = device_status_scope


def test_OnDeviceStatusReceived(vsx_sensor):
    res = vsx_sensor.Connect()
    assert res == 0

    res = vsx_sensor.RegisterOnDeviceStatusReceived(MyOnDeviceStatusReceived)
    assert res == 0

    res = vsx_sensor.SubscribeToDeviceStatusData()
    assert res == 0

    time.sleep(1)
    global global_status_item_list
    print(global_status_item_list)

    res = vsx_sensor.UnsubscribeToDeviceStatusData()
    assert res == 0


@pytest.mark.skip(reason="Only developer testing")
def test_init(vsx_sensor):
    res = vsx_sensor.Connect()
    assert res == 0

    ret = vsx_sensor.GetCurrentDeviceInformation()

    a, b, c = vsx_sensor.TestSystem("VersionId", "CPU")

    res = vsx_sensor.ResetDynamicContainerGrabberEx(5, VsxProtocolDriver.Strategy.DROP_OLDEST)
    assert res == 0

    # res = vsx_sensor.UploadData(r"C:\tmp\0301.bmp")
    # assert res == 0

    input_value = f"""
    <TestCall Name="CaptureData">
        <Parameter Name="SendImage" Value="True"/>
        <Parameter Name="Illumination" Value="Onboard"/>
        <Parameter Name="DataFormat" Value="LeftRaw"/>
        <Parameter Name="ExposureTimeMs" Value="10"/>
        <Parameter Name="CameraGain" Value="50"/>
    </TestCall>
    """

    res = vsx_sensor.TestSystem("CaptureData", input_value)
    assert res[0] == 0

    res, dch = vsx_sensor.GetDataContainer(10000)
    assert res == 0
    res = dch.GetTagList()
    res = dch.GetImage("Image")
    res = dch.SaveData("*", r"c:\tmp\save_dc3.zip")
    dch = None

    res = vsx_sensor.UploadData(r"C:\tmp\0318.bmp")
    assert res == 0

    input_value = f"""
        <TestCall Name="CaptureData">
            <Parameter Name="SendImage" Value="True"/>
            <Parameter Name="Illumination" Value="Onboard"/>
            <Parameter Name="DataFormat" Value="LeftRaw"/>
            <Parameter Name="ExposureTimeMs" Value="10"/>
            <Parameter Name="CameraGain" Value="50"/>
        </TestCall>
        """

    # res = vsx_sensor.TestSystem("CaptureData", input_value)
    # assert res[0] == 0

    res, dch = vsx_sensor.GetDataContainer(10000)
    assert res == 0
    res = dch.SaveData("*", r"c:\tmp\save_dc2.zip")
    dch = None

    vsx_sensor.Disconnect()
    vsx = None


@pytest.fixture
def serial_sensor():
    vsx = VsxProtocolDriver.Sensor.InitSerialSensor("COM4", 115200, "SMARTRUNNER",
                                                    VsxProtocolDriver.SerialConnectionType.RS485,
                                                    "")
    yield vsx
    del vsx


@pytest.mark.xfail
def test_get_set_parameter_serial_sensor(serial_sensor):
    res = serial_sensor.Connect()
    assert res == 0

    illumination_time_ms = "981"
    ret = serial_sensor.SetSingleDataValue(1, "General", 1, "0x680001", illumination_time_ms)
    ret, value = serial_sensor.GetSingleDataValue(1, "General", 1, "0x680001")
    assert ret == 0
    assert value == illumination_time_ms

    illumination_time_ms = "1000"
    serial_sensor.SetSingleDataValue(1, "General", 1, "0x680001", illumination_time_ms)
    ret, value = serial_sensor.GetSingleDataValue(1, "General", 1, "0x680001")
    assert ret == 0
    assert value == illumination_time_ms

    res = serial_sensor.Disconnect()
