﻿
#include <PF.VsxProtocolDriver.WrapperNE.h>

#include "Sr3DParameterIdentifierStereo.h"

#include <cstdio>
#include <string>

static void Consume(VsxSystemHandle* vsx);

void RunExample_Sr3DStereo(VsxSystemHandle* vsx)
{
    VsxStatusCode ret;

    VsxParameterList* parameterList = nullptr;

    ret = vsx_GetParameterList(vsx, &parameterList);
    if (ret == VSX_STATUS_SUCCESS)
    {
        printf("--- Device contains the following parameters: ---\n");

        for (int i = 0; i < parameterList->length; ++i)
        {
            const VsxParameter& p = parameterList->parameters[i];

            std::string value;
            switch (p.valueType) {
            case VSX_PARAMETER_VALUE_TYPE_BOOL:
            case VSX_PARAMETER_VALUE_TYPE_INT:
            case VSX_PARAMETER_VALUE_TYPE_LONG:
            case VSX_PARAMETER_VALUE_TYPE_UINT:
            case VSX_PARAMETER_VALUE_TYPE_INT16:
                printf("Id: %s, Name: %s, %i\n", p.parameterId, p.name, p.valueInt);
                break;
            case VSX_PARAMETER_VALUE_TYPE_FLOAT:
            case VSX_PARAMETER_VALUE_TYPE_DOUBLE:
                printf("Id: %s, Name: %s, %f\n", p.parameterId, p.name, p.valueDouble);
                break;
            case VSX_PARAMETER_VALUE_TYPE_STRING:
            case VSX_PARAMETER_VALUE_TYPE_HEXSTRING:
            case VSX_PARAMETER_VALUE_TYPE_BASE64:
            case VSX_PARAMETER_VALUE_TYPE_ENUM:
            case VSX_PARAMETER_VALUE_TYPE_IP:
            case VSX_PARAMETER_VALUE_TYPE_RECTANGLE:
            case VSX_PARAMETER_VALUE_TYPE_QUAD:
            case VSX_PARAMETER_VALUE_TYPE_POINT:
                printf("Id: %s, Name: %s, %s\n", p.parameterId, p.name, p.valueString);
                break;
            case VSX_PARAMETER_VALUE_TYPE_UNKNOWN:
                printf("Id: %s, Name: %s, Unknown parameter format\n", p.parameterId, p.name);
                break;
            }
        }
        vsx_ReleaseParameterList(&parameterList);
    }
    else {
        printf("Unable to receive parameter list: ");
        print_error(ret);
    }


    printf("Press ENTER to start 3D data grabbing.\n");
    getchar();

    // Start autotrigger and enable trigger, see manual for config Id and parameter Id
    ret = vsx_SetSingleParameterValue(vsx, 1, TriggerSource.ConfigId, 1, TriggerSource.ParamId, TriggerSource.Items[0]);
    if (ret != VSX_STATUS_SUCCESS)
    {
        printf("Error setting parameter: ");
        print_error(ret);
        return;
    }
    // Enable trigger
    ret = vsx_SetSingleParameterValue(vsx, 1, TriggerEnabled.ConfigId, 1, TriggerEnabled.ParamId, "1");
    if (ret != VSX_STATUS_SUCCESS)
    {
        printf("Error setting parameter: ");
        print_error(ret);
        return;
    }

    // Set output format to disparity, then you get the data needed from device
    ret = vsx_SetSingleParameterValue(vsx, 1, OutputMode.ConfigId, 1, OutputMode.ParamId, OutputMode.Items[1]);
    if (ret != VSX_STATUS_SUCCESS)
    {
        printf("Error setting parameter: ");
        print_error(ret);
        return;
    }

    // Start listening for 3D data. 3 containers will be buffered internaly
    ret = vsx_ResetDynamicContainerGrabber(vsx, 3, VSX_STRATEGY_DROP_WRITE);

    for (int i = 0; i < 10; ++i)
    {
        Consume(vsx);
    }

    printf("\n");
    printf("Grabbing finished, device disconnected, press any key to exit.");
    getchar();
}


static void Consume(VsxSystemHandle* vsx)
{
    VsxStatusCode ret;
    VsxDataContainerHandle* container = nullptr;

    ret = vsx_GetDataContainer(vsx, &container, 500);
    if (ret == VSX_STATUS_SUCCESS && container != nullptr)
    {
        printf("Grabbed single 3D image\n");
        VsxImage* xLayerImage = nullptr;
        VsxImage* yLayerImage = nullptr;
        VsxImage* zLayerImage = nullptr;

        // now point cloud data is inside of the calibratedA, calibratedB and calibratedC messages
        VsxStatusCode retX = vsx_GetImage(container, "CalibratedA", &xLayerImage);
        VsxStatusCode retY = vsx_GetImage(container, "CalibratedB", &yLayerImage);
        VsxStatusCode retZ = vsx_GetImage(container, "CalibratedC", &zLayerImage);

        if (retX == VSX_STATUS_SUCCESS && xLayerImage != nullptr)
        {
            // process point cloud data, p.e., or save it with Save3DPointCloudData
            int xWidth = xLayerImage->width;
            int xHeight = xLayerImage->height;
            int xLinePitch = xLayerImage->linePitch;
            if (xLayerImage->format == VSX_IMAGE_DATA2_FORMAT_COORD3D_A32F) // The stereo x image format, y is B32f, ...
            {

                // set the (u, v) coordinates to center of image
                int u = xHeight / 2;
                int v = xWidth / 2;

                // For the SR3D-stereo:
                // xLayerImage is 2 dimensional image buffer containing all the X-values of the 3D point cloud.
                // - Each pixel is a float (4 bytes).
                // - The dimensions of the 2D buffer are described by the members 'width' and 'height' (unit: pixels)
                // - Each line has a size of 'linePitch' BYTES. where linePitch >= width * sizeof(float)

                // the correct way to access the buffer data is as follows:
                char* pixelValuesX = reinterpret_cast<char*>(xLayerImage->rawdata);
                float X = *reinterpret_cast<float*>((pixelValuesX + v * xLinePitch + u * sizeof(float)));
            }

            vsx_ReleaseImage(&xLayerImage);
        }
        if (retY == VSX_STATUS_SUCCESS && yLayerImage != nullptr)
        {
            vsx_ReleaseImage(&yLayerImage);
        }
        if (retZ == VSX_STATUS_SUCCESS && zLayerImage != nullptr)
        {
            vsx_ReleaseImage(&zLayerImage);
        }
        vsx_ReleaseDataContainer(&container);
    }
    else {
        printf("Error grabbing image: ");
        print_error(ret);
    }
}
