﻿using PF.VsxProtocolDriver.Types;
using System.Globalization;

namespace VsxProtocolDriverDemo
{
    public static class SavePointcloudCsv
    {
        public static void Save(IVsxDynamicContainer dynamicContainer, string fileName)
        {
            using (var file = new System.IO.StreamWriter(fileName))
            {
                foreach (var point in GetPoints(dynamicContainer))
                {
                    // write floats to csv file
                    file.WriteLine($"{point.x.ToString("G", CultureInfo.InvariantCulture)}, {point.y.ToString("G", CultureInfo.InvariantCulture)}, {point.z.ToString("G", CultureInfo.InvariantCulture)}");
                }
            }

        }

        private static IEnumerable<(float x, float y, float z)> GetPoints(IVsxDynamicContainer dynamicContainer)
        {
            // now point cloud data is inside of the calibratedA, calibratedB and calibratedC messages
            var xLayerImage = dynamicContainer.GetMessage("CalibratedA") as IVsxImage;
            var yLayerImage = dynamicContainer.GetMessage("CalibratedB") as IVsxImage;
            var zLayerImage = dynamicContainer.GetMessage("CalibratedC") as IVsxImage;

            if (xLayerImage == null)
                throw new ArgumentException("xLayerImage is null");

            if (yLayerImage == null)
                throw new ArgumentException("yLayerImage is null");

            if (zLayerImage == null)
                throw new ArgumentException("zLayerImage is null");

            int xWidth = xLayerImage.Width;
            int xHeight = xLayerImage.Height;
            int xLinePitch = xLayerImage.LinePitch;
            var xPixelFormat = xLayerImage.Format;

            int yWidth = yLayerImage.Width;
            int yHeight = yLayerImage.Height;
            int yLinePitch = yLayerImage.LinePitch;
            var yPixelFormat = yLayerImage.Format;

            int zWidth = zLayerImage.Width;
            int zHeight = zLayerImage.Height;
            int zLinePitch = zLayerImage.LinePitch;
            var zPixelFormat = zLayerImage.Format;

            // The data is organized as 3 planes with 2 dimensions
            // One plane for X-data, Y-data and Z-data
            // Make sure all planes have the same geometry
            if (xWidth != yWidth || xWidth != zWidth)
                throw new ArgumentException("Image geometries are not equal");
            if (xHeight != yHeight || xHeight != zHeight)
                throw new ArgumentException("Image geometries are not equal");
            if (xLinePitch != yLinePitch || xLinePitch != zLinePitch)
                throw new ArgumentException("Image geometries are not equal");

            // SR3D-TOF has 16-bit signed data
            if (xPixelFormat == ImageData2Format.Coord3D_A16 && yPixelFormat == ImageData2Format.Coord3D_B16 && zPixelFormat == ImageData2Format.Coord3D_C16)
            {
                byte[] xData = xLayerImage.ImageData;
                byte[] yData = yLayerImage.ImageData;
                byte[] zData = zLayerImage.ImageData;
                int xInvalidDataValue = (int)xLayerImage.InvalidDataValue;
                int yInvalidDataValue = (int)yLayerImage.InvalidDataValue;
                int zInvalidDataValue = (int)zLayerImage.InvalidDataValue;


                for (int row = 0; row < xHeight; ++row)
                {
                    for (int col = 0; col < xWidth; ++col)
                    {
                        int index = row * xLinePitch + col * sizeof(Int16);
                        // get the x, y, z - value
                        int x = BitConverter.ToInt16(xData, index);
                        int y = BitConverter.ToInt16(yData, index);
                        int z = BitConverter.ToInt16(zData, index);
                        // check if the pixel is invalid
                        if (x != xInvalidDataValue && y != yInvalidDataValue && z != zInvalidDataValue)
                            yield return ((float)x, (float)y, (float)z);
                    }
                }
                yield break;
            }
            // SR3D-Stereo has float data
            if (xPixelFormat == ImageData2Format.Coord3D_A32f && yPixelFormat == ImageData2Format.Coord3D_B32f && zPixelFormat == ImageData2Format.Coord3D_C32f)
            {
                float[] xData = xLayerImage.ImageDataFloats;
                float[] yData = yLayerImage.ImageDataFloats;
                float[] zData = zLayerImage.ImageDataFloats;
                float xInvalidDataValue = (float)xLayerImage.InvalidDataValue;
                float yInvalidDataValue = (float)yLayerImage.InvalidDataValue;
                float zInvalidDataValue = (float)zLayerImage.InvalidDataValue;

                for (int row = 0; row < xHeight; ++row)
                {
                    for (int col = 0; col < xWidth; ++col)
                    {
                        int index = row * xWidth + col;

                        float x = xData[index];
                        float y = yData[index];
                        float z = zData[index];
                        // check if the pixel is invalid
                        if (x != xInvalidDataValue && y != yInvalidDataValue && z != zInvalidDataValue && !float.IsNaN(x) && !float.IsNaN(y) && !float.IsNaN(z))
                            yield return (x, y, z);
                    }
                }
                yield break;
            }
            throw new ArgumentException("Invalid or inconsistent pixel formats");
        }
    }
}
