﻿import VsxProtocolDriver
import Sr3DParameterIdentifierStereo


def main():
    # First discover devices via UDP and use the ip of the first device found
    ret, sensors = VsxProtocolDriver.Sensor.GetUdpDeviceList()

    if ret == VsxProtocolDriver.StatusCode.VSX_STATUS_SUCCESS and len(sensors) > 0:
        # create a new VsxProtocolDriver instance
        dev = sensors[0]
        print("Device found: {0}({1})".format(dev["sensorType"], dev["ipAddress"]))
        sensor = VsxProtocolDriver.Sensor.InitTcpSensor(dev["ipAddress"], "")

    else: # use fix ip address
        # create a new VsxProtocolDriver instance with fix ip address
        sensor = VsxProtocolDriver.Sensor.InitTcpSensor("192.168.2.4", "")

    # Connect with device
    ret = sensor.Connect()
    if ret != VsxProtocolDriver.StatusCode.VSX_STATUS_SUCCESS:
        print(f"Unable to connect with device: {VsxProtocolDriver.Sensor.GetErrorText(ret)}")
        return

    # Get the current device information
    ret, deviceInformation = sensor.GetDeviceInformation()

    if ret == VsxProtocolDriver.StatusCode.VSX_STATUS_SUCCESS:
        print("Connected with device {0}, IP: {1}".format(deviceInformation["sensorType"],
                                                          deviceInformation["ipAddress"]))
        print()
    else:
        print(f"Unable to receive device information: {VsxProtocolDriver.Sensor.GetErrorText(ret)}")
        return

    sensorType = deviceInformation["sensorType"]

    if sensorType == "SR3D_STEREO":
        run_example(sensor)
    else:
        print("Wrong type of sensor connected for this demo")

    # Disconnect device
    sensor.Disconnect()


def run_example(sensor: VsxProtocolDriver.Sensor):
    # Get complete parameter list
    res, param_list = sensor.GetParameterList()
    if res == VsxProtocolDriver.StatusCode.VSX_STATUS_SUCCESS:
        print("--- Device contains the following parameters: ---")
        for p in param_list:
            print(f"Id: {p.parameterId}, Name: {p.name}, Value: {p.value}")
    else:
        print(f"Unable to receive parameter list: {res}")
        return

    print()
    print("Press <RETURN> to grab 20 3D images.")
    input()

    # Start autotrigger and enable trigger, see manual for config Id and parameter Id
    ret = sensor.SetSingleParameterValue(1, Sr3DParameterIdentifierStereo.TriggerSource["ConfigId"], 1,
                                         Sr3DParameterIdentifierStereo.TriggerSource["ParamId"],
                                         Sr3DParameterIdentifierStereo.TriggerSource["Items"][0])

    if ret != VsxProtocolDriver.StatusCode.VSX_STATUS_SUCCESS:
        print(f"Error enabling auto trigger mode: {VsxProtocolDriver.Sensor.GetErrorText(ret)}")
        return

    ret = sensor.SetSingleParameterValue(1, Sr3DParameterIdentifierStereo.TriggerEnabled["ConfigId"], 1,
                                         Sr3DParameterIdentifierStereo.TriggerEnabled["ParamId"], "1")
    if ret != VsxProtocolDriver.StatusCode.VSX_STATUS_SUCCESS:
        print(f"Error enabling trigger: {VsxProtocolDriver.Sensor.GetErrorText(ret)}")
        return

    # Alternative:
    #param = next((p for p in param_list if p.configId == Sr3DParameterIdentifierStereo.TriggerEnabled["ConfigId"]
    #              and p.parameterId == Sr3DParameterIdentifierStereo.TriggerEnabled["ParamId"]))
    #param.value = True
    #ret = sensor.SetSingleParameter(param)

    # Set output format to disparity, then you get the data needed from device
    ret = sensor.SetSingleParameterValue(1, Sr3DParameterIdentifierStereo.OutputMode["ConfigId"], 1,
                                         Sr3DParameterIdentifierStereo.OutputMode["ParamId"],
                                         Sr3DParameterIdentifierStereo.OutputMode["Items"][1])
    if ret != VsxProtocolDriver.StatusCode.VSX_STATUS_SUCCESS:
        print(f"Error setting output format: {VsxProtocolDriver.Sensor.GetErrorText(ret)}")
        return


    # Start listening for 3D data. 3 containers will be buffered internally
    ret = sensor.ResetDynamicContainerGrabberEx(3, VsxProtocolDriver.Strategy.DROP_OLDEST)
    if ret != VsxProtocolDriver.StatusCode.VSX_STATUS_SUCCESS:
        print(f"Error setting up grabber: {VsxProtocolDriver.Sensor.GetErrorText(ret)}")
        return

    for i in range(20):
        consume(sensor)


def consume(sensor: VsxProtocolDriver.Sensor):

    ret, container = sensor.GetDataContainer(500)
    if ret == VsxProtocolDriver.StatusCode.VSX_STATUS_SUCCESS and container is not None:
        # now point cloud data is inside of the calibratedA, calibratedB and calibratedC messages
        xret, xLayerImage, xresults = container.GetImage("CalibratedA")
        yret, yLayerImage, yresults = container.GetImage("CalibratedB")
        zret, zLayerImage, zresults = container.GetImage("CalibratedC")

        if xLayerImage is not None and xresults is not None:
            # process point cloud data, p.e., or save it with Save3DPointCloudData
            xWidth = xresults["width"]
            xHeight = xresults["height"]
            xLinePitch = xresults["line_pitch"]
            pixelValuesX = xLayerImage

            print("3D data grabbed")
        else:
            print("No data received")
    else:
        print(f"Grabbing error: {VsxProtocolDriver.Sensor.GetErrorText(ret)}")


if __name__ == "__main__":
    main()

